#coding=utf-8

from buildz import Base
from buildz.db import tls
from buildz import xf
def dict2list(data):
    rst = []
    for k,v in data.items():
        if type(v) not in (list, tuple):
            v = [v]
        tmp = [k]+list(v)
        rst.append(tmp)
    return rst
def deal_item(item, conf):
    key = item[0]
    if key in "sql_before,sql_after,sql_delete_before,sql_delete_after".split(","):
        conf[key].append(item[1])
    else:
        conf[key] = item[1]
def deal_key(it, conf):
    data_keys = xf.g(conf, data_keys=[])
    tmp = {}
    #print(f"[TESTZ] conf: {conf}, it: {it}")
    for i in range(min(len(it), len(data_keys))):
        tmp[data_keys[i]] = it[i]
    #print(f"[TESTZ] tmp: {tmp}")
    sql_key = tmp['sql_key']
    sql_def = tmp['sql_def']
    py_key = xf.g(tmp, py_key=None)
    if py_key is not None:
        xf.g(conf, py2sqls={})[py_key] = sql_key
    xf.g(conf, keys=[]).append(sql_key)
    xf.g(conf, vars=[]).append(sql_key+" "+sql_def)
def makes(datas, data_keys = "sql_key,sql_def,py_key".split(",")):
    if type(datas)==dict:
        rst = [make(v, k, py_name = k) for k,v in datas.items()]
    else:
        rst = [make(it) for it in datas]
    return rst
def make(data, table=None, data_keys = "sql_key,sql_def,py_key".split(","), py_name=None):
    #print(f"[TESTZ] MARK: {data}")
    """
        从配置中生成建表语句，删表语句和orm对象
        输入:
            data: dict or list
                dict: {
                    (其他配置)
                    (其他配置)
                    表字段: [sql字段定义, 转py后字段名]
                    表字段: sql字段定义
                }
                list: [
                    ((其他配置))
                    ((其他配置))
                    [表字段, sql字段定义, 转py后字段名]
                    [表字段, sql字段定义]
                ]
                其他配置:
                    格式: key, val
                    key:
                        table: 表名
                        query_keys: 默认null，orm做插入/更新时，判断数据是插入还是更新的查询方式
                        auto_translate: 默认false，如果没有配置“转py后字段名”，是否自动转(a_b=>aB)
                        sql_before: 新增一条建表前sql语句
                        sql_after: 新增一条建表后sql语句
                        sql_delete_before: 新增一条删表前sql语句
                        sql_delete_after: 新增一条删表后sql语句
                        data_keys: 默认[sql_key, sql_def, py_key]，非配置项怎么解析，不建议改
            table: 表名
        输出:
            [[sql_create, sql_delete], orm_obj]
    """
    if xf.is_args(data):
        confs = data.lists
        confs =[[it] for it in confs]
        data = dict2list(data.maps)
        data = confs+data
    elif type(data) == dict:
        data = dict2list(data)
    conf = {}
    xf.s(conf, sql_before = [], sql_after = [], sql_delete_before=[], sql_delete_after=[],auto_translate=False,table=table,keys=[],py2sqls={},query_keys=None, data_keys = data_keys, vars=[], py_name = py_name)
    for it in data:
        if len(it)==1 and type(it[0]) in (list, tuple):
            deal_item(it[0], conf)
        else:
            deal_key(it, conf)
    table = xf.g(conf, table=table)
    py_name = xf.g(conf, py_name = py_name)
    if py_name is None:
        py_name = table
    vars =  xf.g(conf, vars = [])
    vars = ",".join(vars)
    assert table is not None
    sql = f"create table if not exists {table}({vars})"
    sql_del = f"drop table if exists {table}"
    sqls = xf.g(conf, sql_before=[], sql_after=[], sql_delete_before=[], sql_delete_after=[])
    #sqls = [";".join(k) for k in sqls]
    sqls_crt = sqls[0]+[sql]+sqls[1]
    sqls_del = sqls[2]+[sql_del]+sqls[3]
    sqls = [sqls_crt, sqls_del]
    sqls = [(";\n".join(k)+";").replace(";;",";") for k in sqls]
    obj = TableObject(*xf.g(conf, keys=[], table=table, py2sqls = None, query_keys=None,auto_translate=False), sql_create = sqls[0], sql_delete = sqls[1], py_name = py_name)
    return sqls, obj

pass

class TableObject(Base):
    def clone(self):
        obj = TableObject(self.keys, self.table, self.src_py2sqls, self.query_keys, self.auto_translate, self.dv, self.sql_create, self.sql_delete, self.py_name)
        return obj
    """
        字段映射
        keys: 表字段
        table: 表名
        py2sqls: 自定义映射
        query_keys: 唯一索引（判断数据重复）
        auto_translate:
            对于没有在py2sqls自定义映射的表字段，是否自动转换(py|autoTran<=>auto_tran|sql)
    """
    def init(self, keys, table=None, py2sqls=None, query_keys=None, auto_translate=False, dv = None, sql_create=None, sql_delete=None, py_name = None):
        self.src_py2sqls = py2sqls
        if table is None:
            table = tls.lower(self.__class__.__name__)
        self.table = table
        if py_name is None:
            py_name = table
        self.py_name = py_name
        if py2sqls is None:
            py2sqls = {}
        if type(py2sqls) in (list, tuple):
            py2sqls = {k[0]:k[1] for k in py2sqls}
        sql2pys = {}
        for k_py, k_sql in py2sqls.items():
            sql2pys[k_sql] = k_py
        for key in keys:
            if key not in sql2pys:
                k_py, k_sql = key,key
                if auto_translate:
                    k_py,k_sql = tls.upper(key), tls.lower(key)
                py2sqls[k_py] = k_sql
                sql2pys[k_sql] = k_py
        self.py2sqls = py2sqls
        self.sql2pys = sql2pys
        self.keys = keys
        if query_keys is None:
            query_keys = []
        self.query_keys = query_keys
        self.dv = dv
        self.py2sql=self.to_sql
        self.sql2py = self.to_py
        self.sql_create = sql_create
        self.sql_delete = sql_delete
        self.auto_translate = auto_translate
    def create(self, dv=None):
        assert self.sql_create is not None
        if self.sql_create.find("<table>")>=0:
            print(f"[WARN] not create template table by {self.sql_create}")
            return
        #print(f"[TESTZ] sql_create: {self.sql_create}")
        dv = self.rdv(dv)
        dv.executes(self.sql_create)
    def drop(self, dv=None):
        assert self.sql_delete is not None
        if self.sql_delete.find("<table>")>=0:
            print(f"[WARN] not delete template table by {self.sql_delete}")
            return
        self.rdv(dv).execute(self.sql_delete)
    def bind(self, dv):
        self.dv = dv
    def to_sql(self, obj):
        if type(obj)!=dict:
            tmp = {}
            for k in self.py2sqls:
                if hasattr(obj, k):
                    tmp[k] = getattr(obj, k)
            obj = tmp
        obj = {self.py2sqls[k]:v for k,v in obj.items() if k in self.py2sqls}
        return obj
    def to_sql_keys(self, keys):
        obj = [k for k in keys if k in self.py2sqls]
        return obj
    def to_py(self, obj):
        if type(obj)!=dict:
            tmp = {}
            for k in self.sql2pys:
                if hasattr(obj, k):
                    tmp[k] = getattr(obj, k)
            obj = tmp
        obj = {self.sql2pys[k]:v for k,v in obj.items()}
        return obj
    def rdv(self, dv):
        if dv is None:
            dv = self.dv
        assert dv is not None
        return dv
    def query(self, sql, dv=None, sql2py=False):
        dv=self.rdv(dv)
        rst = dv.query(sql, as_map=1)
        if sql2py:
            rst = [self.sql2py(it) for it in rst]
        return rst
    def query_all(self, dv=None, sql2py=False):
        sql = f"select * from {self.table};"
        return self.query(sql, dv, sql2py)
    def execute(self, sql, dv=None):
        dv=self.rdv(dv)
        return dv.execute(sql)
    def filter_sql(self, obj):
        if type(obj)!=dict:
            tmp = {}
            for k in self.sql2py:
                if hasattr(obj, k):
                    tmp[k] = getattr(obj, k)
            obj = tmp
        obj = {k:v for k,v in obj.items() if k in self.keys}
        return obj
    def commit(self, dv=None):
        dv = self.rdv(dv)
        dv.execute("commit")
    def save(self, obj, dv=None, py2sql = True, commit=False, check = True, update_keys = None):
        dv = self.rdv(dv)
        if type(update_keys)==str:
            update_keys = [update_keys]
        if type(obj) not in [list, tuple]:
            obj = [obj]
        if py2sql:
            obj = [self.to_sql(k) for k in obj]
            if update_keys is not None:
                update_keys = self.to_sql_keys(update_keys)#[self.to_sql_keys(k) for k in update_keys]
        else:
            obj = [self.filter_sql(k) for k in obj]
        query_keys = self.query_keys
        if not check:
            query_keys = None
        _ = [dv.insert_or_update(k, self.table, query_keys, update_keys) for k in obj]
        if commit:
            dv.execute("commit")
        return _
    def clean(self, dv=None,commit = False):
        dv = self.rdv(dv)
        dv.execute(f"delete from {self.table}")
        if commit:
            dv.execute("commit")
        





#coding=utf-8

import hashlib
import os
#计算文件hash值
def fhash(fp, hm="sha256", blk_sz = 10240):
	if type(hm) == str:
		hm = getattr(hashlib, hm)()
	with open(fp, 'rb') as f:
		while True:
			bs = f.read(blk_sz)
			if len(bs)==0:
				break
			hm.update(bs)
	return hm.hexdigest()

pass
def bs_hash(bs, fhm):
	obj = fhm()
	obj.update(bs)
	return obj.hexdigest()

pass
def fhashs(fp, blk_size = 10240, max_read = -1, hm = "sha256"):
	if type(hm) == str:
		fhm = getattr(hashlib, hm)
	else:
		fhm = hm
	hashs = []
	if not os.path.isfile(fp) or max_read==0:
		return hashs
	size = 0
	bs = b""
	reads = 0
	mark_read = True
	with open(fp, 'rb') as f:
		while len(bs)>0 or mark_read:
			if max_read>0 and size>=max_read:
				break
			read_size = blk_size
			if max_read>0:
				read_size = min(max_read-size, blk_size)
			if mark_read:
				curr = f.read(read_size)
				if len(curr)==0:
					mark_read = False
				else:
					bs += curr
			if len(bs)>=read_size or not mark_read:
				tmp = bs[:read_size]
				bs = bs[read_size:]
				size+=read_size
				hashs.append(bs_hash(tmp, fhm))
	return hashs
	



#coding=utf-8
import sys
import os
curr_dir = os.path.dirname(__file__)
mark_msvc = False
mark_mingw32 = False
if len(sys.argv)==1:
    sys.argv.append("build_ext")
    #sys.argv.append("--inplace")
    sys.argv.append('--build-lib')
    sys.argv.append(f"{curr_dir}")
if sys.platform=='win32':
    mark_msvc = True
    for k in sys.argv[1:]:
        if k.find("--compiler=")==0:
            v=k[len("--compiler="):].strip().lower()
            mark_msvc = v=='msvc'
            mark_mingw32 = v=='mingw32'
            break
from distutils.core import setup, Extension
from Cython.Build import cythonize
extra_link_args = []
define_macros=[]
extra_compile_args = []
sources = [os.path.join(curr_dir, "pcxf.pyx"),os.path.join(curr_dir, "pc.cpp")]
from buildz import fz
cpps = fz.search(os.path.join(curr_dir,"loaderz"),".*\.cpp")
sources+=cpps
if not mark_msvc:
    extra_link_args = ["-lstdc++", "-O3"]
    extra_compile_args = ["-O3"]
else:
    extra_compile_args = ["/O2"]
if mark_mingw32:
    define_macros = [("MS_WIN64",1)]
    from setuptools._distutils import cygwinccompiler
    msvcrs = cygwinccompiler.get_msvcr()
    if "ucrt" in msvcrs:
        # windows下没有预装ucrt，但python编译要ucrt，直接把mingw相关的库静态打包进去，就不用ucrt了
        cygwinccompiler.get_msvcr = lambda :[]
        extra_link_args.append("-static")
#extra_link_args.append("-static")

setup(
    ext_modules = cythonize(Extension(
        "pcxf",
        language="c++",
        sources = sources,
        include_dirs=[os.path.join(curr_dir,"loaderz"),curr_dir],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,  # 链接 C++ 标准库
        define_macros=define_macros,
    )),
)




'''
单元测试注解
'''
from ..base import Base,fcBase
from .. import logz
import traceback
class UnitTest(Base):
    def init(self, log=None):
        self.tests = []
        self.log = logz.make(log)
    def case(self, key):
        key = str(key)
        def f(fc):
            self.tests.append([key, fc])
        return f
    def call(self, key = None, args=[], maps={}):
        if key is None:
            key = ''
        rst = []
        self.log.info(f"start unit test on key: {key}")
        for _key, fc in self.tests:
            if _key.find(key)==0:
                self.log.info("    unit test case: ", _key)
                try:
                    fc(*args, **maps)
                except exp:
                    import traceback
                    self.log.error("    unit test exp: ", exp)
                    self.log.error(traceback.format_exc())
                self.log.info("    done unit test case: ", _key)

pass
def build(*a,**b):
    return UnitTest(*a,**b)



import inc, time, os
from multiprocessing import Process
import subprocess
print('test x')
def fc():
    while True:
        print(f"subprocess: {os.getpid()} running")
        time.sleep(1.0)
def test():
    try:
        inc.fc()
        print("xxxtestxxxxxxxx", __name__)
        return
        subprocess.run("python inc.py")
        print("done subprocess")
        p = Process(target=fc, daemon=True)
        print("process")
        p.start()
        print("process start")
        while True:
            print(f"process {os.getpid()} running")
            time.sleep(0.5)
            #p.kill()
    finally:
        pass
pass



from ..base import *
basepath = path
path = pathz.Path()
path.set("conf", basepath.local("ioc/conf"), curr=0)
class Encape(Base):
    def call(self, params=None, **maps):
        return None

pass
class Deal(Base):
    def deal(self, conf, unit):
        return None
    def update(self, conf, unit):
        return conf, False
    def call(self, conf, unit):
        'encape, conf, conf_need_udpate'
        conf, upd = self.update(conf, unit)
        encape = self.deal(conf,unit)
        return encape,conf,upd

pass

class Params(Base):
    def str(self):
        return f"iocz.Params(args={self.args}, maps={self.maps})"
    @staticmethod
    def Clone(params, **upds):
        if params is None:
            args,maps = [],{}
        else:
            args, maps = list(params.args), dict(params.maps)
        maps.update(upds)
        obj = Params()
        obj.args = args
        obj.maps = maps
        return obj
    def clone(self, **upds):
        return Params.Clone(self, **upds)
    def init(self, *args, **maps):
        self.args = args
        self.maps = maps
    def set(self, key, val):
        self.maps[key] = val
    def get(self, key, default=None):
        if key not in self.maps:
            return default
        return self.maps[key]
    def __getattr__(self, key):
        if key not in self.maps:
            return None
        return self.maps[key]

pass



from .base import *
from ... import dz
class Conf(Base):
    '''
        list to dict
    '''
    def init(self, v2l=True):
        self.lists = {}
        self.aliases = {}
        self.ranges = {}
        self.v2l = v2l
    @staticmethod
    def default(val):
        if val is None:
            val = [0, None]
        if not dz.islist(val):
            val = [1, val]
        return val
    def range(self, key, base, last=None, min = 0, default=None):
        self.ranges[key] = base,last,min,self.default(default)
        return self
    def ikey(self, index=None, key=None, aliases=[], need=False, remove=True, deal = None, dict_out = False, default=None):
        if index is not None:
            self.index(index, key, need, deal, dict_out, default)
        if key is not None:
            self.key(key, aliases, need, remove, deal, dict_out, default)
        return self
    def key(self, key, aliases=[], need=False, remove=True, deal = None, dict_out = False, default=None):
        self.aliases[key] = list(aliases), need, remove, deal, dict_out, self.default(default)
        return self
    def index(self, i, key=None, need=False, deal = None, dict_out = False, default=None):
        self.lists[i] = key,need,deal, dict_out, self.default(default)
        return self
    def range_to_dict(self, rst, conf, unit=None):
        for key, item in self.ranges.items():
            base,last,min,default = item
            if last is None:
                tmp = conf[base:]
            else:
                tmp = conf[base:last]
            if len(tmp)<min:
                if default[0]:
                    val = default[1]
                else:
                    assert 0,f"require key '{key}' not set in list {conf}"
            else:
                val = tmp
            rst[key] = val
    def to_dict(self, conf, unit=None):
        if not dz.islist(conf):
            if not dz.isdict(conf) and self.v2l:
                conf = [conf]
            else:
                return conf, False
        rst = {}
        for i,item in self.lists.items():
            key, need, deal, out_dict, default = item
            if i>=len(conf):
                if default[0]:
                    val = default[1]
                else:
                    assert not need, f"require key '{key}' not set in list {conf}"
                    continue
            else:
                val = conf[i]
            if deal is not None:
                val,upd = deal(val, unit)
            if type(val)==dict and out_dict:
                rst.update(val)
            else:
                rst[key] = val
        self.range_to_dict(rst, conf, unit)
        return rst, 1
    def update_dict(self, conf, unit=None):
        upd = False
        for key, item in self.aliases.items():
            aliases, need, remove, deal, dict_out, default = item
            if key not in conf:
                for name in aliases:
                    if name in conf:
                        conf[key] = conf[name]
                        upd = 1
                        if remove:
                            del conf[name]
                        break
            if key not in conf:
                if default[0]:
                    conf[key] = default[1]
                else:
                    assert not need, f"require key '{key}' not set in dict {conf}"
            if key in conf and deal is not None:
                conf[key], _upd = deal(conf[key], unit)
                upd = upd or _upd
            if key in conf:
                val = conf[key]
                if dz.isdict(val) and dict_out:
                    dz.fill(val, conf, 1)
        return upd
    def call(self, conf, unit=None):
        conf, upd = self.to_dict(conf, unit)
        upd = upd or self.update_dict(conf, unit)
        return conf, upd

pass




#

import ssl
from . import mhttp

def load_verify_context(cafile=None, capath=None, cadata=None,check_hostname=True):
    if cafile is not None or capath is not None or cadata is not None:
        # 导入根证书，需要单个根证书文件cafile或者根证书文件夹capath或者根证书数据cadata
        # cadata是证书数据，不清楚能不能多个证书字节码拼在一起
        srv_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        srv_context.load_verify_locations(cafile=cafile, capath=capath, cadata=cadata)
        srv_context.check_hostname = check_hostname
    else:
        # 不会校验服务端https证书是否有效，可能有风险？
        srv_context = ssl._create_unverified_context(ssl.PROTOCOL_TLS_CLIENT)
    return srv_context

pass

def load_server_context(fp_cert, fp_prv, pwd=None):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(fp_cert, fp_prv, password=pwd)
    return context

def wrap(context, wrap_cli, server_side, is_wrap = True):
    if is_wrap:
        wrap_cli = wrap_cli.skt
    return mhttp.WSocket.Bind(context.wrap_socket(wrap_cli, server_side=server_side))



from . import buffer
from . import base
from . import pos
from . import exp
from . import item
from ..stack import Stack
class Manager:
    def __init__(self, as_bytes = False):
        self.index = 0
        self.as_bytes = as_bytes
        if as_bytes:
            self.type = bytes
        else:
            self.type = str
        self.deals = {}
        self.builds = {}
        self.default_deal = self.like("")
    def default_deals(self):
        return self.deals[self.default_deal]
    def get_deals(self, c):
        if c in self.deals:
            return self.deals[c]
        return []
    def like(self, s):
        if self.type ==type(s):
            return s
        if type(s)==str:
            return s.encode()
        return s.decode()
    def add(self,obj):
        obj.regist(self)
        lbs = obj.labels()
        lbs = [self.like(k[:1]) for k in lbs]
        for lb in lbs:
            if lb not in self.deals:
                self.deals[lb] = []
            self.deals[lb].append(obj)
        types = obj.types()
        for _type in types:
            if _type not in self.builds:
                self.builds[_type] = []
            self.builds[_type].append(obj)
        return self
    def do(self, fcs, *argv, **maps):
        for fc in fcs:
            if fc(*argv, **maps):
                return True
        return False
    def regist(self):
        id = self.index
        self.index+=1
        return "id_"+str(id)
    def build(self, obj):
        if obj.is_val:
            return obj
        #print(f"[TESTZ] obj: {obj}")
        _type = obj.type
        if _type not in self.builds:
            raise exp.Exp(f"unspt type: {_type}", obj.pos)
        builds = self.builds[_type]
        for deal in builds:
            rst = deal.build(obj)
            if rst is not None:
                return rst
        raise exp.Exp(f"unspt deal type:[{obj}]", obj.pos)
    def deal(self, buffer, arr):
        c = buffer.read()
        if len(c)==0 and buffer.size()==0:
            return False
        deals = self.get_deals(c)
        find = False
        for deal in deals:
            if deal.deal(buffer, arr, self):
                return True
        deals = self.default_deals()
        for deal in deals:
            if deal.deal(buffer, arr, self):
                return True
        return False
    def build_arr(self, _arr):
        dts = []
        for k in _arr:
            _k = self.build(k)
            if item.is_null(_k):
                continue
            dts.append(_k)
        return dts
    def load(self, buffer):
        try:
            return self._load(buffer)
        except exp.Exp as _exp:
            #import traceback
            #traceback.print_exc()
            exp.deal(_exp, buffer)
    def _load(self, buffer):
        arr = []
        while self.deal(buffer, arr):
            pass
        arr = self.build_arr(arr)
        arr_pos = base.arr_pos(arr)#(arr[0].pos[0], arr[-1].pos[-1])
        #print(f"mg arr: {arr}")
        obj = item.Item(arr, arr_pos, type = "list", is_val = 0)
        obj = self.build(obj)
        #arr = rst
        #if len(arr)==1:
        #    arr = arr[0]
        val = obj.val
        #print(f"mg val:{val}")
        if type(val) in [list,map] and len(val)==0:
            val = ""
        if type(val)==list and len(val)==1:
            val = val[0]
        return val
    def loads(self, reader):
        if type(reader) == self.type:
            #print(f"try str, {len(reader)}")
            buff = buffer.StrBuffer(reader)
        else:
            buff = buffer.Buffer(reader)
        return self.load(buff)

pass



#coding=utf-8
from buildz import xf, argx, pyz, ioc, fz,tz

from ..test import text_sfx, args, maps
class Deal:
    def deal(self):
        if len(args)<5:
            print("need opt, file1, file2, step_file")
            return
        opt = args[1]
        fp1 = args[2]
        fp2 = args[3]
        if len(args)>4:
            fp_step = args[4]
        else:
            fp_step = "steps.diff"
        mark_encode = argx.get(maps, 'e', 1)
        mark_txt = argx.get(maps, 't', None)
        if mark_txt:
            mark_encode = 0
        spt = argx.get(maps, 's', 1)
        nspt = argx.get(maps, 'ns', 0)
        if nspt:
            spt = 0
        if opt == 'diff' or opt == 'count':
            bs1 = fz.read(fp1).decode("utf-8")
            bs2 = fz.read(fp2).decode("utf-8")
            #bs1 = b"text z xxxyzijsa"
            #bs2 = b"test xxxx afzjcovijsax"
            stps = tz.m_steps(bs1, bs2,split=spt)
            if opt == 'count':
                cnt = tz.m_count(stps)
                print(f"diff from {fp1} to {fp2}: {cnt}")
                return
            if mark_encode:
                stps = b"e"+tz.m_encode(stps)
            else:
                for stp in stps:
                    c = stp[-1]
                    if type(c)==bytes:
                        stp[-1] = c.decode()
                stps = "t"+xf.dumps(stps)
                stps = stps.encode()
            fz.write(stps, fp_step)
            print("done diff")
        elif opt == 'update':
            bs1 = fz.read(fp1).decode("utf-8")
            bs_step = fz.read(fp_step)
            mark_encode = bs_step[:1] == b'e'
            bs_step= bs_step[1:]
            if mark_encode:
                stps = tz.m_decode(bs_step)
            else:
                stps = xf.loads(bs_step.decode())
            bs2 = tz.m_update(bs1, stps,split=spt).encode("utf-8")
            fz.write(bs2, fp2)
            print("done update")
        else:
            print(f"unexpect opt: {opt}")
    pass

pass



from buildz import Base, xf, pathz
import os
dp = os.path.dirname(__file__)
path = pathz.Path()
path.set("res", os.path.join(dp, "res"))
class Eval(Base):
    def call(self, obj):
        return True
class Build(Base):
    def init(self):
        self.builder = None
    def bind(self, builder):
        self.builder = builder
class FcEval(Eval):
    def init(self, fc, eval):
        self.eval = eval
        self.fc = fc
    def call(self, obj):
        return self.fc(self.eval(obj))
    @staticmethod
    def Make(fc):
        def f(eval):
            return FcEval(fc, eval)
        return f
class FcBuild(Build):
    def init(self, fc):
        self.fc = fc
    def call(self, data):
        data = data['data']
        data = self.builder(data[0])
        return self.fc(data)
    @staticmethod
    def Make(fc):
        def f():
            return FcBuild(fc)
        return f
class ListEval(Eval):
    def init(self, evals = None):
        if evals is None:
            evals = []
        self.evals = evals
    def add(self, eval):
        self.evals.append(eval)
    def list_call(self, obj):
        assert 0, "not impl"
    def call(self, obj):
        assert len(self.evals)>0, "empty list evals"
        return self.list_call(obj)
class FcListEval(ListEval):
    def init(self, fc, evals=None, num = -1):
        super().init(evals)
        self.fc = fc
        self.num = num
    def list_call(self, obj):
        if self.num >= 0:
            assert len(self.evals)==self.num
        rst = self.evals[0](obj)
        for eval in self.evals[1:]:
            val = eval(obj)
            rst = self.fc(rst, val)
        return rst
    @staticmethod
    def Make(fc, num=-1):
        def f(evals=None):
            return FcListEval(fc, evals, num)
        return f
class ListBuild(Build):
    def list_call(self, arr):
        assert 0, "not impl"
    def call(self, data):
        data = data['data']
        arr = [self.builder(it) for it in data]
        return self.list_call(arr)
class FcListBuild(ListBuild):
    def init(self, fc, num=-1):
        super().init()
        self.num =num
        self.fc = fc
    def list_call(self, arr):
        if self.num>=0:
            assert self.num==len(arr)
        return self.fc(arr)
    @staticmethod
    def Make(fc, num=-1):
        def f():
            return FcListBuild(fc, num)
        return f
AndEval = FcListEval.Make(lambda x,y:x and y)
AndBuild = FcListBuild.Make(AndEval)
OrEval = FcListEval.Make(lambda x,y:x or y)
OrBuild = FcListBuild.Make(OrEval)
XorEval = FcListEval.Make(lambda x,y:x != y)
XorBuild = FcListBuild.Make(XorEval)
EqualEval = FcListEval.Make(lambda x,y:x == y)
EqualBuild = FcListBuild.Make(EqualEval)
BiggerEval = FcListEval.Make(lambda x,y:x > y)
BiggerBuild = FcListBuild.Make(BiggerEval)
BiggerEqualEval = FcListEval.Make(lambda x,y:x >= y)
BiggerEqualBuild = FcListBuild.Make(BiggerEqualEval)
LittlerEval = FcListEval.Make(lambda x,y:x < y)
LittlerBuild = FcListBuild.Make(LittlerEval)
LittlerEqualEval = FcListEval.Make(lambda x,y:x<=y)
LittlerEqualBuild = FcListBuild.Make(LittlerEqualEval)
PatternEval = FcListEval.Make(lambda val, pt:len(re.findall(pt, val))>0, num=2)
PatternBuild = FcListBuild.Make(PatternEval, 2)
InEval = FcListEval.Make(lambda val, lst: val in lst, num=2)
class InBuild(Build):
    def call(self, data):
        data = data['data']
        assert len(data)==2
        val = self.builder(data[0])
        arr = [self.builder(it) for it in data[1]]
        return InEval([val, arr])
NotEval = FcEval.Make(lambda x: not x)
NotBuild = FcBuild.Make(NotEval)
class ValEval(Eval):
    def init(self, val):
        self.val = val
    def call(self, obj):
        return self.val
class ValBuild(Build):
    def call(self, data):
        data = data['data'][0]
        return ValEval(data)
default_builds = xf.loadf(path.res("default.js"))
class EvalBuilder(Base):
    """
        {
            type: and|or|xor|not|=|>|<|pt|...
            data: [
                {type: var, data: name}
                {type: val, data: 'test'}
            ]
        }
        eq(var(name),val(test))
    """
    def init(self, default_type = None):
        self._default_type = default_type
        self.builds = {}
        self._default_build = None
    def default_type(self, dtype):
        self._default_type = dtype
    def default_build(self, fc):
        self._default_build = fc
        return self
    def set(self, key, fc):
        if type(key) not in (list, tuple):
            key = [key]
        fc.bind(self)
        for k in key:
            self.builds[k] = fc
        return self
    def call(self, data):
        if type(data) not in (list, tuple, dict):
            data = [self._default_type, data]
        if type(data) in (list, tuple):
            data = {'type':data[0], 'data':data[1:]}
        if type(data['data']) not in (list, tuple, set):
            data['data'] = [data['data']]
        tp = data['type']
        if tp not in self.builds and self._default_build is not None:
            return self._default_build(data)
        assert tp in self.builds
        return self.builds[tp](data)
    @staticmethod
    def Make(arr=[], default_sets = True, default_type = 'const'):
        obj = EvalBuilder(default_type)
        if default_sets:
            for k, opts in default_builds.items():
                fc = globals()[k]()
                if type(opts) not in (list, tuple, set):
                    opts = [opts]
                for opt in opts:
                    obj.set(opt, fc)
        for fc, opts in arr:
            if type(opts) not in (list, tuple, set):
                opts = [opts]
            for opt in opts:
                obj.set(opt, fc)
        return obj

pass


from . import itemz
class BaseDeal:
    def height_tree(self, mg, obj):
        return [1]
    def deal_encape(self, rst, conf, deep, height, up_height, left, right, top_not_format):
        left = conf.s(left)
        right = conf.s(right)
        num_spc = conf.get(line=4)
        spc = conf.s(conf.get(spc=' '))
        curr_spc = spc*(num_spc*(deep-1))
        spt = conf.s(self.spt)
        if not conf.check(format=1) or height<=conf.get(nfmt_height=1):
            spc = conf.s(" ")*conf.get(prev=0)
            spc = spt+spc
            rst = left+spc.join(rst)+right
            rst = self.prev_spc(conf, deep, up_height, top_not_format)+rst
        else:
            enter = conf.s("\n")
            rst = (spt+enter).join(rst)
            rst = left+enter+rst+enter+curr_spc+right
            if not top_not_format:
                rst = curr_spc+rst
        return rst
    def prev_spc(self, conf, deep, up_height, top_not_format):
        nfmt_height = conf.get(nfmt_height=1)
        if top_not_format:
            return conf.s("")
        if conf.check(format=1) and up_height>nfmt_height:
            num_spc = conf.get(line=4)
            spc = conf.s(conf.get(spc=' '))
            curr_spc = spc*(num_spc*(deep-1))
            return curr_spc
        return conf.s("")
    def fmt(self, s, num, spc = " "):
        spc = self.like(spc, s)
        spc = spc*num
        et = self.like("\n", s)
        arr = s.split(et)
        arr = [spc+k for k in arr]
        rs = et.join(arr)
        return rs
    def deal(self, mg, obj, up_height, heights, conf, deep, top_not_format):
        "impl"
        assert 0
    def like(self, s, target):
        if type(target) ==type(s):
            return s
        if type(s)==str:
            return s.encode()
        return s.decode()
    def __str__(self):
        return "BaseDeal"
    def __repr__(self):
        return str(self)
    def __init__(self, *argv, **maps):
        self.init(*argv, **maps)
    def init(self, *argv, **maps):
        pass


#
# set PYTHONVERBOSE=1
import pcxf
def build_str(bts):
    return bts.decode("utf-8")
def deal_exp(s):
    print(s)

pass
def test():
    import time
    n_list =1500
    n_dict=500
    n=3
    #n_arr = 10
    obj = [1,2,3]
    rst = []
    for i in range(n_list):
        rst.append(obj)
    obj = rst
    rst = {}
    for j in range(n_dict):
        rst[j] = obj
    obj = rst
    import json
    from buildz import xf
    from buildz.xf import readz_nexp
    s = json.dumps(obj)
    bs = s.encode("utf-8")
    print("start")
    for i in range(n):
        curr=time.time()
        jobj = json.loads(s)
        sec = time.time()-curr
        print(f"json loads: {sec}, {type(jobj)}")
    print(f"xf.loads:", xf.loads)
    for i in range(n):
        curr=time.time()
        rst = xf.loads(bs)
        sec = time.time()-curr
        print("xf rst:", type(rst), "sec:", sec)
    # print(f"readz_nexp.loads:", readz_nexp.loads)
    # for i in range(n):
    #     curr=time.time()
    #     rst = readz_nexp.loads(bs)
    #     sec = time.time()-curr
    #     print("readz_nexp rst:", type(rst), "sec:", sec)
    for i in range(n):
        curr=time.time()
        tst_obj = pcxf.loads(bs)
        sec = time.time()-curr
        print(f"pcxf loads: {sec}, {type(tst_obj)}, {tst_obj==jobj}")
rst = pcxf.loads(b"{1={1=2,3=4,'123'=(1,2,3)},3=4,'5'=[1,2,3]}")
#rst = pcxf.py_loads("{}", build_str, deal_exp)
print("rst:",rst)
test()




from buildz import argz,xf,Base,evalz,pathz
import sys,os

path = pathz.Path()
path.set("res", path(path.dir(__file__), "tests"))
conf = xf.loadf(path.res("conf.js"))

def search(dp, filepath = None, content = None):
    print(f"call in search: ({dp}, {filepath}, {content})")
    return f"call in search {dp}"

pass


mg = argz.CallBuilder().var("search", search)
fc = mg(conf)

ins = xf.args(base=0)
args, maps = ins.args, ins.maps

rst = fc(args, maps)
print(f"rst: {rst}")
'''
python -m buildz.argz.test search /d/rootz
python -m buildz.argz.test search /d/rootz f=asf c=test
'''


#
from ..ioc.base import Base, EncapeData
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class EnvDeal(FormatDeal):
    """
        环境变量env:
            {
                id: id
                type: env
                env|key: 环境变量key
                default: item_conf //可选
            }
        简写：
            [[id, env], key, default]
            [env, key, default]
            [env, key]
        例:
            [env, path] //读取环境变量path
    """
    def init(self, fp_lists=None, fp_defaults=None):
        super().init("EnvDeal", fp_lists, fp_defaults, join(dp, "conf", "env_lists.js"), None)
    def deal(self, edata:EncapeData):
        data = edata.data
        data = self.fill(data)
        key = xf.get_first(data, 'env', 'key')
        val = edata.conf.get_env(key)
        default = xf.g(data, default=None)
        if val is None and default is not None:
            val = self.get_obj(default, edata.conf)
        return val

pass



from .. import base
from .. import item
from .. import exp
from . import lr
import re
class ValDeal(base.BaseDeal):
    def has_prev(self):
        return 0
    def has_deal(self):
        return 1
    """
        正则表达式匹配
    """
    def init(self, pt, fc):
        st = self.like("^", pt)
        ed = self.like("$", pt)
        if pt[0]!=st:
            pt = st+pt
        if pt[-1]!=ed:
            pt = pt+ed
        self.pt = pt
        self.fc = fc
    def deal(self, queue, stack):
        if len(queue)==0:
            return False
        _item = queue[0]
        if not _item.check(is_val=1, type = None):
            return False
        val = _item.val
        if type(val) not in [bytes, str]:
            return False
        pt = self.like(self.pt, val)
        if re.match(pt, val) is None:
            return False
        try:
            val = self.fc(val)
        except Exception as exp1:
            print("exp:", exp1)
            raise exp.FormatExp("error in reval fc:\""+str(exp1)+"\"", _item.pos, _item.val)
        queue.pop(0)
        out_item = item.DealItem(val, _item.pos, self.id(), is_val = 1)
        stack.append(out_item)
        return True

pass


#coding=utf-8
from . import pyz
def fc_s2l(s):
    if type(s)!=str:
        return s
    arr = s.split(",")
    arr = [k.strip() for k in arr]
    return arr
def fcBase(cls, lst=0, keys = tuple(), fcs = tuple()):
    keys = fc_s2l(keys)
    fcs = fc_s2l(fcs)
    class _Base(cls):
        def str(self):
            return str(self.__class__)
        def repr(self):
            return self.str()
        def __str__(self):
            return self.str()
        def __repr__(self):
            return self.repr()
        # def __call__(self, *a, **b):
        #     return self.call(*a, **b)
        # def call(self, *a, **b):
        #     return None
        def init(self, *a, **b):
            pass
        def __init__(self, *a, **b):
            sa = a[:lst]
            a = a[lst:]
            sb = {}
            for k in keys:
                if k in b:
                    sb[k] = b[k]
                    del b[k]
            super().__init__(*sa, **sb)
            self.init(*a, **b)
    for fc in fcs:
        base_fc = "__"+fc+"__"
        def _base_fc(obj, *a, **b):
            return getattr(obj, fc)(*a, **b)
        setattr(_Base, base_fc, _base_fc)
    return _Base

class Base:
    def str(self):
        return str(self.__class__)
    def __str__(self):
        return self.str()
    def __repr__(self):
        return self.__str__()
    def __init__(self, *args, **maps):
        self.init(*args, **maps)
    def __call__(self, *args, **maps):
        return self.call(*args, **maps)
    def init(self, *args, **maps):
        pass
    def call(self, *args, **maps):
        return self.deal(*args, **maps)
    def deal(self, *args, **maps):
        return None

pass

class WBase(Base):
    def _open(self):
        pass
    def open(self):
        self._open()
        return pyz.with_out(self.close, True)
    def _close(self):
        pass
    def close(self, exc_type, exc_val, exc_tb):
        self._close()

pass

class With(Base):
    def init(self, args=False):
        self.args = args
    def call(self, cls):
        def _open(obj):
            pass
        def open(obj):
            obj._open()
            return pyz.with_out(obj.close, self.args)
        def _close(obj):
            pass
        def close(obj):
            obj._close()
        cls._open = _open
        cls._close = _close
        cls.open = open
        cls.close = close
        return cls

pass

class Args(Base):
    @staticmethod
    def is_args(obj):
        return isinstance(obj, Args)
    @staticmethod
    def is_collect(obj):
        return Args.is_args(obj) or type(obj) in (list, tuple, dict)
    def __len__(self):
        return self.size()
    def as_dict(self, deep=False):
        mps = dict(self.dicts)
        if deep:
            rst = {}
            for k,v in mps.items():
                if isinstance(k, Args):
                    k = k.as_dict(deep)
                if isinstance(v, Args):
                    v = v.as_dict(deep)
                rst[k]=v
            mps = rst
        for it in self.lists:
            if isinstance(it, Args):
                if len(it.lists)!=len(it):
                    continue
                it = it.lists
            if len(it)!=2:
                continue
            if deep:
                it = [k if not isinstance(k, Args) else k.as_dict(deep) for k in it]
            mps[it[0]] = it[1]
        return mps
    def as_list(self, cmb=0, item_args = True, deep=False):
        lst = list(self.args)
        if deep:
            lst = [it if not isinstance(it, Args) else it.as_list(cmb, item_args, deep) for it in lst]
        mps = dict(self.dicts)
        for k,v in mps.items():
            if deep and isinstance(v, Args):
                v = v.as_list(cmb, item_args, deep)
            if isinstance(k, Args):
                k = k.as_list(cmb, item_args, deep)
            if cmb==0 or type(v)!=list:
                it = [k,v]
            else:
                v = list(v)
                v.append(k)
            if item_args:
                it = Args(v)
            lst.append(it)
        return lst
    def size(self):
        return len(self.args)+len(self.maps)
    def str(self):
        return f"<Args args={self.args}, maps={self.maps}>"
    @property
    def lists(self):
        return self.args
    @lists.setter
    def lists(self, val):
        self.args=val
    @lists.deleter
    def lists(self):
        del self.args
    @property
    def dicts(self):
        return self.maps
    @dicts.setter
    def dicts(self, val):
        self.maps=val
    @dicts.deleter
    def dicts(self):
        del self.maps
    def init(self, args, maps):
        self.args = args
        self.maps = maps

pass


#
from .base import *
from .datas import Datas
from .confs import Confs
from .encapes import Encapes
from ... import pyz

class Builds(Base):
    def init(self, unit):
        self.need_build = False
        self.jobs = []
        self.dones = []
        self.unit = unit
        self.builds = None
    def bind(self, builds):
        if self.builds == builds:
            return
        self.builds = builds
        self.builds.add(self)
    def add(self, conf):
        self.jobs.append(conf)
        self.need_build = True
        if self.builds is not None:
            self.builds.add(self)
    def build(self):
        if not self.need_build:
            return
        self.need_build = False
        for conf in self.jobs:
            if Confs.is_conf(conf):
                encape,c,u = self.unit.get_encape(conf, self.unit.ns, self.unit.id)
            elif isinstance(conf, Encape):
                encape = conf
            else:
                encape,c,u = self.unit.get_encape(conf, self.unit.ns, self.unit.id)
            encape()
            self.dones.append(conf)
        self.jobs = []
    def call(self):
        self.build()
class Buildset(Base):
    def init(self, mg):
        self.mg = mg
        self.need_build = False
        self.jobs = []
        self.dones = []
    def add(self, builds):
        if builds.need_build:
            self.jobs.append(builds)
            self.need_build = True
        builds.bind(self)
    def build(self):
        if not self.need_build:
            return
        self.need_build = False
        for builds in self.jobs:
            builds()
            self.dones.append(builds)
        self.jobs = []
        self.need_build = False
    def call(self):
        self.build()

pass



import sys
import os
import hashlib
import inspect
#from .base import Base, WBase
class With:
    def __init__(self, fc_in, fc_out, args = False):
        self.fc_in = fc_in
        self.fc_out = fc_out
        self.args = args
    def __enter__(self):
        if self.fc_in is not None:
            self.fc_in()
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        rst = False
        if self.fc_out is not None:
            if self.args:
                rst = self.fc_out(exc_type, exc_val, exc_tb)
            else:
                rst = self.fc_out()
        return rst

pass
def with_out(fc, args = False):
    return With(None, fc, args)

pass
def hashcode(s):
    if type(s)==str:
        s = s.encode("utf-8")
    return hashlib.md5(s).hexdigest()

pass
def add_prevdir(fp, up = 1, adds = []):
    """
        将当前文件所在目录的上{up}层目录加入sys.path，这样可以在同层写测试代码，但是写import的时候还是要当作是在上{up}层做import
    """
    if type(adds) not in [list, tuple]:
        adds = [adds]
    fp = os.path.abspath(fp)
    if os.path.isfile(fp):
        dp = os.path.dirname(fp)
    else:
        dp = fp
    for i in range(up):
        dp = os.path.dirname(dp)
    dp = os.path.join(dp, *adds)
    sys.path.insert(0, dp)

pass
add_prev_dir = add_prevdir
test_current = add_prevdir
add_path = test_current
add_current = test_current
add = test_current
def load(md, fc = None):
    """
        import object(whether module or others) from md(or md.fc)
        exp:
            load("buildz.xf") = package xf
            load("buildz.xf", "loads") = function loads from package buildz.xf
            load("buildz.xf.loads") = function loads from package buildz.xf
    """
    mds = md.split(".")
    arr = mds[1:]
    while len(mds)>0:
        try:
            md = __import__(".".join(mds))
            break
        except ModuleNotFoundError as exp:
            mds = mds[:-1]
    if len(mds)==0:
        raise Exception("can't import package from "+md)
    for k in arr:
        md = getattr(md, k)
    if fc is not None:
        fc = getattr(md, fc)
    else:
        fc = md
    return fc

pass
def pyexe():
    return sys.executable

pass
exe=pyexe
is_windows = sys.platform.lower()=='win32'
def pypkg():
    """
        return python package path, test on linux and windows
    """
    try:
        import buildz
        import os
        return os.path.dirname(buildz.__path__[0])
    except:
        pass
    import site
    sites = site.getsitepackages()
    if is_windows:
        fpath = sites[-1]
    else:
        fpath = sites[0]
    return fpath

pass
pkg = pypkg
pth = pypkg

class Pth:
    def __init__(self, fp = "build.pth"):
        self.fp = os.path.join(pth(), fp)
    def read(self):
        fp = self.fp
        if not os.path.isfile(fp):
            return []
        with open(fp, 'rb') as f:
            s = f.read().decode()
        return s.split("\n")
    def add(self, path):
        arr = self.read()
        if path in arr:
            print("alread add")
            return
        arr.append(path)
        self.write(arr)
    def write(self, paths = []):
        if type(paths) not in [list, tuple]:
            paths = [paths]
        s = "\n".join(paths)
        with open(self.fp, 'wb') as f:
            f.write(s.encode())
    def remove(self):
        os.remove(self.fp)

pass

_pth = Pth()
def main(name, fc, *args, **maps):
    if name=="__main__":
        fc(*args, **maps)

pass

def bylocals(mlocals, fc, *args, **maps):
    if mlocals['__name__']=="__main__":
        fc(*args, **maps)

pass
#lcmain=bylocals
lc_main=bylocals
lc = bylocals
local = bylocals
def mainerr(fc, *args, **maps):
    st = inspect.stack()[1]
    if st.filename == '<stdin>':
        fc(*args, **maps)
    else:
        print("not main:", st.filename)

pass

def not_null(*args, exp_def=None):
    for v in args:
        if v is not None:
            return v
    return ret_exp_def(exp_def)
nnull = not_null
class exp_def_enp:
    def __init__(self, exp=False, default=None, exp_fc = Exception, exp_args = [], exp_maps = {}):
        self.exp_fc = exp_fc
        self.exp_args = exp_args
        self.exp_maps=exp_maps
        self.exp = exp
        self.default = default
    def __call__(self, exp=None, default=None, exp_fc = None, exp_args=None, exp_maps=None):
        exp = not_null(exp, self.exp)
        default = not_null(default, self.default)
        exp_fc = not_null(exp_fc, self.exp_fc)
        exp_args=not_null(exp_args, self.exp_args)
        exp_maps = not_null(exp_maps, self.exp_maps)
        if exp:
            raise exp_fc(*exp_args, **exp_maps)
        return default
pass
def default(val):
    return exp_def_enp(False, val)

exp_def = exp_def_enp
def ret_exp_def(val):
    if isinstance(val, exp_def_enp):
        return val()
    return val

def decode(s, coding = 'utf-8'):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return s.decode(coding)
    except:
        return s.decode(xcoding)

pass
def encode(bs, coding="utf-8"):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return bs.encode(coding)
    except:
        return bs.encode(xcoding)


def fc_input(s, fc_done):
    s = input(s)
    fc_done(s)

pass
def th_input(s, fc_done):
    import threading as th
    t = th.Thread(target=fc_input, args=(s, fc_done), daemon=True)
    t.start()

pass


#

import sys
from buildz.gpuz.torch import DictCache
from buildz.gpuz.test import analyze
from buildz import pyz
import math
import torch,time
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
cpu,cuda = analyze.dvs
MultiheadAttention = nn.MultiheadAttention
class PostionalEncoding(nn.Module):
    def __init__(self, word_size, max_length, batch_first=False):
        super().__init__()
        pos = torch.arange(max_length).unsqueeze(1)
        words = torch.arange(word_size)
        mod2 = words%2
        offset = torch.pi*0.5*mod2
        # exp(log(A)*B)等价于A^B
        x = pos*torch.exp(-math.log(1e4)*(words-mod2)/word_size)
        data = torch.sin(x+offset)
        data.requires_grad=False
        if batch_first:
            data = data.unsqueeze(0)
        else:
            data = data.unsqueeze(1)
        self.vecs = nn.Parameter(data, requires_grad=False)
        self.batch_first = batch_first
    def forward(self, ins):
        if self.batch_first:
            ins = ins + self.vecs[:, :ins.size(1)]
        else:
            ins = ins + self.vecs[:ins.size(1)]
        return ins
def make_linear(input_size, middle_size, middle_fc):
    if middle_size is not None:
        nets = []
        nets.append(nn.Linear(input_size, middle_size))
        if middle_fc is not None:
            nets.append(middle_fc())
        nets.append(nn.Linear(middle_size, input_size))
        linear = nn.Sequential(*nets)
    else:
        linear = nn.Linear(input_size, input_size)
    return linear
class Decoder(nn.Module):
    def __init__(self, word_dims, num_heads=1, kv_dims=None, batch_first=False, bias=True, linear_size=None, linear_fc = nn.ReLU):
        super().__init__()
        self.self_multi = MultiheadAttention(word_dims,num_heads,bias=bias,batch_first=batch_first,kdim=kv_dims,vdim=kv_dims)
        self.linear = make_linear(word_dims, linear_size, linear_fc)
        self.ln1 = nn.LayerNorm(word_dims)
        self.ln2 = nn.LayerNorm(word_dims)
        self.batch_first = batch_first
    def forward(self, outputs, outs_mask):
        self_attn, _ = self.self_multi(outputs,outputs,outputs, attn_mask=outs_mask)
        outputs = self.ln1(outputs+self_attn)
        outputs = self.ln2(outputs+self.linear(outputs)) 
        return outputs
class DecodePart(nn.Module):
    def __init__(self, num_words, word_dims, sequence_length, num_decoders=6, num_heads=1, kv_dims=None, batch_first=False, bias=True, linear_size=None, linear_fc = nn.ReLU):
        super().__init__()
        self.embedding = nn.Embedding(num_words, word_dims)
        self.pos_encoding = PostionalEncoding(word_dims, sequence_length, batch_first)
        self.num_decoders = num_decoders
        decoders = [Decoder(word_dims, num_heads, kv_dims, batch_first, bias, linear_size, linear_fc) for i in range(num_decoders)]
        self.src_decoders = decoders
        self.decoders = nn.ModuleList(decoders)
    def mds(self):
        return [self.embedding, self.pos_encoding]+self.src_decoders 
    def forward(self, outputs, outs_mask):
        outputs = self.embedding(outputs)
        outputs = self.pos_encoding(outputs)
        for decoder in self.decoders:
            outputs = decoder(outputs, outs_mask)
        return outputs
class Chats(nn.Module):
    def __init__(self, num_words_outputs, word_dims, sequence_length_outputs, num_decoders=6, num_heads=1, kv_dims=None, batch_first=False, bias=True, linear_size=None, linear_fc = nn.ReLU, mask_index = -1):
        super().__init__()
        self.decode = DecodePart(num_words_outputs, word_dims, sequence_length_outputs, num_decoders, num_heads, word_dims, batch_first, bias, linear_size, linear_fc)
        self.linear = nn.Linear(word_dims, num_words_outputs)
        self.sequence_length_outputs = sequence_length_outputs
        self.mask_index = mask_index
        self.batch_first = batch_first
        self.num_heads = num_heads
    def mds(self):
        return [self.linear]+self.decode.mds()
    def gen_masks(self, outputs):
        if self.batch_first:
            batch_id, dt_id=0,1
        else:
            batch_id, dt_id=1,0
        batch_size, outs_len = outputs.size(batch_id), outputs.size(dt_id)
        mask = torch.ones(batch_size, self.num_heads, outs_len, outs_len)
        mask = torch.triu(mask, diagonal=1)
        outs_mask = mask.to(torch.bool)
        if self.mask_index>=0:
            outs_mask_1 = (outputs==self.mask_index).unsqueeze(1).unsqueeze(1)
            if not self.batch_first:
                outs_mask_1 = outs_mask_1.transpose(0,3)
            if self.num_heads>1 or outs_len>1:
                outs_mask_1 = outs_mask_1.expand(batch_size, self.num_heads, outs_len, outs_len)
            outs_mask = outs_mask | outs_mask_1
        outs_mask = outs_mask.reshape(-1, outs_len, outs_len).bool().to(outputs.device)
        return outs_mask
    def forward(self, outputs):
        outs_mask = self.gen_masks(outputs)
        rst = self.decode(outputs, outs_mask)
        rst = self.linear(rst)
        return rst

class TestDataset(Dataset):
    def __init__(self, num, std_len, words):
        self.num = num
        self.datas = torch.randint(0, words, (num,std_len))
        sz = analyze.sz(self.datas)
        sz, unit = analyze.show_size(sz)
        print(f"data size: {sz} {unit}")
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.datas[i]

pass

def test():
    num_words_outputs = 1024
    word_dims = 512
    sequence_length_outputs = 512
    num_decoders=12
    num_heads=8
    loop = 5
    num_datas = 60
    batch=30
    lr=0.0001
    win_size=3
    args = sys.argv[1:]
    mark_train = True
    if len(args)>0:
        mark_train = args.pop(0).lower()=='train'
    modes = 'cuda,cache,cpu'
    if len(args)>0:
        modes = args.pop(0)
    if len(args)>0:
        num_decoders = int(args.pop(0))
    print(f"num_decoders: {num_decoders}")
    ds = TestDataset(num_datas, sequence_length_outputs, num_words_outputs)
    dl = DataLoader(ds, batch)
    def fc_gen():
        gmodel = Chats(num_words_outputs, word_dims, sequence_length_outputs, num_decoders, num_heads, batch_first=1)
        mds = gmodel.mds()
        opts =[optim.Adam(md.parameters(), lr=lr) for md in mds]
        gopt = optim.Adam(gmodel.parameters(), lr=lr)
        return mds, gmodel, opts, gopt
    def fc_opt(net, opt):
        torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
        opt.step()
    loss_fn = nn.CrossEntropyLoss()
    def wrap_fn(outs, targets):
        outs = outs.view(-1, outs.shape[-1])
        targets = targets.reshape(-1)
        return loss_fn(outs, targets)
    analyze.analyzes(mark_train, loop, fc_gen, dl, wrap_fn, fc_opt, win_size, modes)

pyz.lc(locals(),test)



from buildz.ioc import wrap
x = wrap.ns("test")
@wrap.obj(id='x')
class A:
    pass

pass
@x.obj(id='y')
class B:
    pass

pass
from buildz import ioc
mg= ioc.build()
x = mg.get('x')
y = mg.get('test.y')
print(x)
print(y)




#coding=utf-8
from buildz.base import Base
from buildz import xf
from threading import Lock
class Fcs(Base):
    def init(self, wrap, fcs):
        self._ioc_args = [wrap, fcs]
        self._base = wrap
    @property
    def conf(self):
        return self._ioc_args[0].obj
    def __getattr__(self, key):
        return self._ioc_args[1][key]

pass
class IOCDObj(Base):
    def init(self):
        self._wrap = None
    def bind(self, wrap):
        self._wrap = wrap
        return self
    @property
    def wrap(self):
        return self._wrap
    @property
    def decorator(self):
        return self._wrap

pass
class IOCDWrap(Base):
    def init(self, fc, wrap):
        self.wrap = wrap
        while isinstance(fc, IOCDWrap):
            fc = fc.fc
        self.fc = fc
    def call(self, *args, **maps):
        fc = self.fc(*args, **maps)
        if isinstance(fc, IOCDObj):
            fc.bind(self.wrap)
        return fc
        #return self.fc(*args, **maps).bind(self.wrap)

pass
class Decorator(Base):
    """
        用@备注的形式自动生成conf文档
    """
    @property
    def namespace(self):
        return self.ns
    def init(self, ns=None, lock = False):
        self.conf = {}
        self.ns = ns
        self.fcs = {}
        self.obj_fcs = Fcs(self, self.fcs)
        self.obj = None
        self.id = None
        self.bind_fcs = []
        self.do_lock = lock
        if lock:
            self.lock = Lock()
            self.set = self.lk_fc(self.set)
            self.get = self.lk_fc(self.get)
            self.add = self.lk_fc(self.add)
            self.add_bind = self.lk_fc(self.add_bind)
    def add_bind(self, fc):
        fc = IOCDWrap(fc, self)
        self.bind_fcs.append(fc)
    def lk_fc(self, fc):
        def _fc(*a, **b):
            with self.lock:
                return fc(*a, **b)
        return _fc
    def bind(self, confs):
        self.conf['namespace'] = self.ns
        #from buildz import xf
        #print(f"[TESTZ] Decorator conf: {self.conf}")
        self.id = confs.add(self.conf)
        self.obj = confs.get_conf(self.id)
        for fc in self.bind_fcs:
            fc()
    def call(self):
        return self.obj_fcs
    def regist(self, key, fc):
        fc = IOCDWrap(fc, self)
        self.fcs[key]=fc
    def clone(self, ns=None, lock=None):
        if lock is None:
            lock = self.do_lock
        dc = Decorator(ns, lock)
        for k,fc in self.fcs.items():
            dc.regist(k, fc)
        for fc in self.bind_fcs:
            dc.add_bind(fc)
        return dc
    def get(self, tag, index):
        conf = self.conf
        if tag not in conf:
            conf[tag]=[]
        return conf[tag][index]
    def add(self, tag, data):
        conf = self.conf
        if tag not in conf:
            conf[tag]=[]
        id = len(conf[tag])
        conf[tag].append(data)
        return id
    def set(self, tag, key, val):
        conf = self.conf
        if tag not in conf:
            conf[tag]={}
        conf[tag][key]=val
    def add_datas(self, item):
        if type(item)==str:
            item = xf.loads(item)
        return self.add("datas", item)
    def get_datas(self, id):
        return self.get("datas", id)
    def set_datas(self, id, val):
        return self.set("datas", id, val)
    def set_envs(self, key, val):
        return self.set("env", key, val)
    def add_inits(self, val):
        return self.add("inits", val)
    def add_locals(self, item):
        return self.add("locals", item)

pass

class Decorators(Base):
    decorator_fc_names = "regist,get,add,set,add_datas,get_datas,set_datas,set_envs,add_inits,add_locals".split(",")
    def init(self, ns=None):
        self.demo = Decorator(ns)
        self.ns = ns
        self.confs = {}
        self.confs[ns] = self.demo
        self.obj = None
        for nfc in Decorators.decorator_fc_names:
            setattr(self, nfc, self.ns_fc(nfc))
    def ns_fc(self, nfc):
        def _fc(*a, ns=None):
            return getattr(self.decorator(ns), nfc)(*a)
        return _fc
    def decorator(self, ns=None):
        if ns is None:
            ns = self.ns
        if ns not in self.confs:
            self.confs[ns] = self.demo.clone(ns)
        return self.confs[ns]
    def call(self, ns=None):
        return self.decorator(ns)()
    def bind(self, confs):
        for ns, conf in self.confs.items():
            conf.bind(confs)
        self.obj = confs

pass
decorators = Decorators()
decorator = decorators.demo
ns = decorators
# class Decorator(Base):
#     def init(self):
#         #self.conf = {}
#         self.confs = {}
#         self.confs[None] = {}
#         self.objs = {}
#         self.ids = {}
#         self.namespace = None
#         self.fcs = {}
#         self._ns = {}
#         self.regist("add_datas", self.add_datas)
#     def get_conf_obj(self, ns):
#         return self.objs[ns]
#     @property
#     def conf(self):
#         return self.get_conf_obj(self.namespace)
#     def fcns(self, namespace, fc):
#         self._ns[fc] = namespace
#     def ns(self, namespace):
#         self.namespace = namespace
#     def curr_ns(self):
#         return self.namespace
#     def regist(self, key, fc):
#         self.fcs[key] = fc
#     def get_conf(self, src, ns = None):
#         if ns is None:
#             ns = self.namespace
#         if src in self._ns:
#             ns = self._ns[src]
#         if ns not in self.confs:
#             conf = {}
#             conf['namespace'] = ns
#             self.confs[ns] = conf
#         return self.confs[ns]
#     def get(self, tag, index, src=None):
#         conf = self.get_conf(src)
#         if tag not in conf:
#             conf[tag]=[]
#         return conf[tag][index]
#     def add(self, tag, data, src = None, ns = None):
#         conf = self.get_conf(src, ns)
#         if tag not in conf:
#             conf[tag]=[]
#         id = len(conf[tag])
#         conf[tag].append(data)
#         return id
#     def set(self, tag, key, val, src=None):
#         conf = self.get_conf(src)
#         if tag not in conf:
#             conf[tag]={}
#         conf[tag][key]=val
#     def add_datas(self, item, key=None, ns = None):
#         if type(item)==str:
#             item = xf.loads(item)
#         return self.add("datas", item, key, ns)
#     def get_datas(self, id, key=None):
#         return self.get("datas", id, key)
#     def set_datas(self, id, val):
#         return self.set("datas", id, val)
#     def set_envs(self, key, val):
#         return self.set("env", key, val)
#     def add_inits(self, val):
#         return self.add("inits", val)
#     def add_locals(self, item):
#         return self.add("locals", item)
#     def bind_confs(self, confs):
#         for ns, val in self.confs.items():
#             id = confs.add(val)
#             obj = confs.get_conf(id)
#             self.ids[ns] = id
#             self.objs[ns] = obj
#     def all(self):
#         arr = [val for k,val in self.confs.items()]
#         return arr
#     # def call(self):
#     #     return self.conf

# pass

# decorator = Decorator()
# class Fcs:
#     def __init__(self, k, ns):
#         self._ioc_ns = [k, ns]
#     @property
#     def conf(self):
#         return self._ioc_ns[1].get_conf(self._ioc_ns[0])

# pass
# class NameSpace(Base):
#     def get_conf(self, ns):
#         return self.decorator.get_conf_obj(ns)
#     def init(self, decorator):
#         self.decorator = decorator
#         self.lock = Lock()
#     def fc(self, namespace, rfc):
#         def wfc(*a, **b):
#             with self.lock:
#                 ns = self.decorator.curr_ns()
#                 self.decorator.ns(namespace)
#                 rst = rfc(*a,**b)
#                 self.decorator.fcns(namespace, rst)
#                 self.decorator.ns(ns)
#                 return rst
#         return wfc
#     def call(self, namespace):
#         fcs = self.decorator.fcs
#         obj = Fcs(namespace, self)
#         for k,f in fcs.items():
#             setattr(obj, k, self.fc(namespace, f))
#         def wfc(rfc, *a, **b):
#             with self.lock:
#                 ns = self.decorator.curr_ns()
#                 self.decorator.ns(namespace)
#                 rst = rfc(*a,**b)
#                 self.decorator.fcns(namespace, rst)
#                 self.decorator.ns(ns)
#                 return rst
#         setattr(obj, "wrap", wfc)
#         return obj

# pass
# ns = NameSpace(decorator)



from .. import base
from .. import item
from .. import exp
from . import lr
class ListDeal(lr.LRDeal):
    """
        分隔符，有分隔符后将缓存的数据当作字符串
    """
    def init(self, left, right):
        super().init(left, right, 'list')
    def types(self):
        return ['list']
    def build(self, obj):
        val = obj.val
        if len(val)==0:
            obj.val = []
            obj.is_val = 1
            return obj
        if len(val)>1:
            opt = val[1]
            if opt.type=='kv':
                return None
        return self.build_arr(val, obj.pos)
    def build_arr(self, arr, arr_pos):
        rst = []
        #pos = self.arr_pos(arr)#(arr[0].pos[0],arr[-1].pos[-1])
        for _item in arr:
            if not _item.is_val:
                raise exp.Exp(f"error in list: item is not val: {_item}", _item.pos)
            rst.append(_item.val)
        return item.Item(rst, arr_pos, type='list', is_val = 1)

pass



from .. import base
from .. import itemz
import re
import json
class ValDeal(base.BaseDeal):
    def deal(self, mg, val, up_height, heights, conf, deep, top_not_format):
        val = conf.s(json.dumps(val, ensure_ascii=False))
        return self.prev_spc(conf, deep, up_height, top_not_format)+val
pass 



            

                


#
import torch
from torch import nn
import threading as th
from ... import Base
import numpy as np
'''
注：
没写好，还不能用
'''
class MultiDictCache(Base):
    def init(self, nets, win_sizes=None, backward_deal=None, fc_inputs_to=None, default_dv = None, default_win_size=1):
        '''
            nets: [
                net
                (net, run_dv)
                (net, run_dv, opt)
                (net, (run_dv, cache_dv), opt)
            ]
            win_sizes: {
                run_dv: size
            }
        '''
        if default_dv is None:
            default_dv = torch.device("cpu")
        if fc_inputs_to is None:
            fc_inputs_to = self.inputs_to
        self.fc_inputs_to = fc_inputs_to
        self.default_dv = default_dv
        if win_sizes is None:
            win_sizes = {}
        self.default_win_size = default_win_size
        self.win_sizes = win_sizes
        #run_dvs = {}
        rst = {}
        nears = {}
        for it in nets:
            if type(it) not in (list,tuple):
                it = [it]
            if len(it)==1:
                it.append(default_dv)
            x = it[1]
            if type(x) not in (list,tuple):
                it[1] = [it[1]]
            run_dvs.add(it[1][0])
            net = it[0]
            net_id = id(net0)
            rst[net_id] = it
            run_dv = it[1][0]
            if run_dv not in nears:
                nears[run_dv] = {}
            nears[run_dv][net_id] = [-1,-1]
            run_dvs[run_dv] = it
        nets = rst
        self.nets = nets
        self.nears = nears
        if type(dvs) not in (list,tuple):
            dvs = [dvs]
        self.gdv = dvs[0]
        self.cdv = dvs[-1]
        self.dvs = dvs
        if dvs_nets is None:
            dvs_nets = [[]]
        if type(dvs_nets) not in (list, tuple):
            dvs_nets = [dvs_nets]
        dvs_nets = [k if k is not None else [] for k in dvs_nets]
        ids_nocache = set()
        fcs = []
        for i in range(len(dvs_nets)):
            for net in dvs_nets[i]:
                fc = net.register_forward_pre_hook(self.hook_forward_pre_tensor(i))
                fcs.append(fc)
                ids_nocache.add(id(net))
        self.ids_nocache = ids_nocache
        self.dvs_nets = dvs_nets
        fcs += [net.register_forward_pre_hook(self.hook_forward_pre) for net in nets if id(net) not in ids_nocache]
        fcs+=[net.register_forward_hook(self.hook_forward) for net in nets if id(net) not in ids_nocache]
        net = nets[0]
        if hasattr(net, "register_full_backward_hook"):
            fcs+=[net.register_full_backward_hook(self.hook_backward) for net in nets]
        else:
            fcs+=[net.register_backward_hook(self.hook_backward) for net in nets]
        self.src_nets = nets
        self.fcs = fcs
        self.nets = {id(net):net for net in nets}
        self.ctxs = {id(net):[] for net in nets}
        if opts is not None:
            opts = {id(net): opt for net,opt in zip(nets, opts)}
        self.opts = opts
        self.pools = []
        self.win_size = win_size
        self.backward_deal = backward_deal
        self.nears = {id(net):[-1,-1] for net in nets}
        self.curr = -1
        self.done_backward = -1
        self.no_caches = True
    def remove(self):
        [fc.remove() for fc in self.fcs]
    def hook_forward_pre_tensor(self, i):
        def hook(model, ins):
            model.to(self.dvs[i])
            ins = self.fc_inputs_to(ins, self.dvs[i])
            return ins
        return hook
    @staticmethod
    def inputs_to(inputs, dv):
        if isinstance(inputs, torch.Tensor):
            return inputs.to(dv)
        inputs = tuple([k.to(dv) for k in inputs])
        return inputs
    def ctxs_to(self, i, dv):
        if dv is None:
            self.ctxs[i] = []
        else:
            self.ctxs[i] = [k.to(dv) for k in self.ctxs[i]]
    def copy_backward(self, nid):
        for c_id in self.pools:
            self.nets[c_id].to(self.cdv)
            self.ctxs_to(c_id, None)
        self.pools = []
        self.nets[nid].to(self.gdv)
        self.ctxs_to(nid, self.gdv)
        self.pools.append(nid)
        next_id = nid
        for i in range(self.win_size-1):
            next_id = self.nears[next_id][1]
            if next_id<0:
                break
            self.nets[next_id].to(self.gdv)
            self.ctxs_to(next_id, self.gdv)
            self.pools.append(next_id)
    def copy_forward(self, nid, model):
        for c_id in self.pools:
            self.nets[c_id].to(self.cdv)
            self.ctxs_to(c_id, self.cdv)
        self.pools = []
        model.to(self.gdv)
        self.pools.append(nid)
        next_id = nid
        for i in range(self.win_size-1):
            next_id = self.nears[next_id][0]
            if next_id<0:
                break
            self.nets[next_id].to(self.gdv)
            self.pools.append(next_id)
    def hook_forward(self, model, ins, outs):
        self.no_caches = True
    def hook_forward_pre(self, model, ins):
        self.no_caches = False
        nid = id(model)
        if nid not in self.pools:
            self.copy_forward(nid, model)
        if self.curr>=0:
            self.nears[self.curr][0] = nid
        self.curr = nid
        return self.fc_inputs_to(ins, self.gdv)
    def do_forward(self, fc):
        self.no_caches = True
        self.ctxs = {k:[] for k in self.nets}
        with torch.autograd.graph.saved_tensors_hooks(self.hook_pack, self.hook_unpack):
            rst = fc()
        return rst.to(self.gdv)
    def wrap_backward_deal(self, net_id):
        if self.backward_deal is None:
            return
        self.backward_deal(self.nets[net_id], self.opts[net_id])
    def hook_backward(self, model, grad_ins, grad_outs):
        nid = id(model)
        self.wrap_backward_deal(nid)
    def hook_pack(self, dt):
        if self.no_caches:
            # 不做缓存，数据不处理
            return -1, dt
        index = self.curr
        # forward时候为了后面计算梯度存的缓存，放到列表里方便转cpu和gpu
        self.ctxs[index].append(dt)
        return index, len(self.ctxs[index])-1
    def hook_unpack(self, x):
        nid = x[0]
        if nid<0:
            return x[1]
        if nid not in self.pools:
            self.copy_backward(nid)
        dt = self.ctxs[nid][x[1]]
        if self.curr>=0:
            self.nears[self.curr][1] = nid
        self.curr = nid
        return dt
    def do_backward(self, fc):
        self.curr = -1
        fc()




from . import itemz
from ..readz import is_args
class Manager:
    def __init__(self):
        self.deals = {}
    def height_tree(self, obj):
        tobj = type(obj)
        if tobj not in self.deals:
            raise Exception("undealable type"+str(obj))
        for fc in self.deals[tobj]:
            rst = fc.height_tree(self, obj)
            if rst is not None:
                return rst
        raise Exception("undealable obj in type"+str(obj))
    def add(self, types, fc):
        if type(types) not in (list, tuple):
            types = [types]
        for _type in types:
            if _type not in self.deals:
                self.deals[_type]=[]
            self.deals[_type].append(fc)
    def deal(self, obj, up_height, heights, conf, deep=1, top_not_format = 0):
        tobj = type(obj)
        if tobj not in self.deals:
            raise Exception("undealable type"+str(obj))
        for fc in self.deals[tobj]:
            rst = fc.deal(self, obj, up_height, heights, conf, deep, top_not_format)
            if rst is not None:
                return rst
        raise Exception("undealable obj in type"+str(obj))
    def dump(self, obj, conf):
        heights = self.height_tree(obj)
        rst = self.deal(obj, heights[0]+1, heights, conf)
        return rst

pass


from buildz import pyz
from ..base import Args
def isdict(obj):
    return type(obj)==dict
def islist(obj):
    return type(obj) in (list, tuple)
def tolist(obj):
    if not islist(obj):
        obj = [obj]
    return obj
def dict2list(obj):
    if type(obj)!=dict:
        return obj
    return list(obj.items())
def dict2iter(obj):
    if type(obj)!=dict:
        return obj
    return obj.items()
def get(obj, key, default=None, set = False):
    if type(obj)==list:
        key = int(key)
        if len(obj)<=key:
            return default
        return obj[key]
    else:
        if key not in obj:
            if set:
                obj[key] = default
            return default
        return obj[key]

pass
def getf(obj, key, fc):
    if key not in obj:
        return fc()
    return obj[key]

pass
def gfn(obj, **maps):
    rst = []
    for k in maps:
        if k in obj and obj[k] is not None:
            v = obj[k]
        else:
            v = maps[k]()
        rst.append(v)
    if len(rst)==1:
        rst = rst[0]
    return rst

pass
def gf(obj, **maps):
    rst = []
    for k in maps:
        if k in obj:
            v = obj[k]
        else:
            v = maps[k]()
        rst.append(v)
    if len(rst)==1:
        rst = rst[0]
    return rst

pass
def get_first(obj, keys, default = None):
    for k in keys:
        if k in obj:
            return obj[k],1
    return default,0

pass
def get_one(obj, keys, default=None):
    return get_first(obj, keys, default)[0]
def g1(obj, **maps):
    v = None
    for k in maps:
        v = maps[k]
        if k in obj:
            v = obj[k]
            return v
    return v

pass
def g(obj, **maps):
    rst = []
    if obj is None:
        return maps.values()
    for k in maps:
        v = maps[k]
        if k in obj:
            v = obj[k]
        rst.append(v)
    if len(rst)==1:
        rst = rst[0]
    return rst

pass
def c(obj, **maps):
    for k in maps:
        v = maps[k]
        if k not in obj:
            return False
        vo = obj[k]
        if vo!=v:
            return False
    return True
def gs(obj, **maps):
    rst = []
    for k in maps:
        v = maps[k]
        if k in obj:
            v = obj[k]
        else:
            obj[k] = v
        rst.append(v)
    if len(rst)==1:
        rst = rst[0]
    return rst

pass
def get_set(obj, key, val):
    if key not in obj:
        obj[key]=val
    return obj[key]
def s(obj, **maps):
    for k in maps:
        v = maps[k]
        obj[k] = v

pass
def fill_fc(maps, fc):
    for k in maps:
        v = maps[k]
        if type(v) == dict:
            fill_fc(v, fc)
            continue
        maps[k] = fc(v)
    return maps

pass
def im2l(maps):
    if type(maps)!=dict:
        return maps
    mx = 0
    for k in maps:
        if type(k)==str:
            k = int(k)
        if type(k)!=int or k<0:
            raise Exception(f"args key only can be int and bigger than 0, but {type(k)}:{k} found")
        mx = max(k+1, mx)
    arr = [None]*mx
    for k, v in maps.items():
        arr[k] = v
    return arr

pass
    
def dset(maps, keys, val):
    if type(keys) not in (list,tuple):
        keys = [keys]
    for key in keys[:-1]:
        if type(maps)==dict:
            if key not in maps:
                maps[key] = {}
        else:
            key = int(key)
        maps = maps[key]
    maps[keys[-1]] = val

pass
def dhas(maps, keys):
    if type(keys) not in (list,tuple):
        keys = [keys]
    for key in keys:
        if type(maps)==dict:
            if key not in maps:
                return False
        else:
            key = int(key)
            if key>=len(maps):
                return default
        maps = maps[key]
    return True

pass

def dget(maps, keys, default = None):
    if type(keys) not in (list,tuple):
        keys = [keys]
    for key in keys:
        find = 1
        if type(maps)==dict:
            find = key in maps
        elif type(maps) in (list, tuple):
            try:
                key = int(key)
            except:
                key = len(maps)
            find = key<len(maps)
        else:
            find = False
        if not find:
            return default,0
        maps = maps[key]
    return maps,1

pass
def dremove(maps, keys):
    if type(keys) not in (list,tuple):
        keys = [keys]
    arr = []
    for key in keys:
        if key not in maps:
            break
        arr.append([maps, key])
        maps = maps[key]
    arr.reverse()
    first=1
    for maps,key in arr:
        if first or len(maps[key])==0:
            del maps[key]
        first=0
    return

pass

def dhget(maps, keys):
    if type(keys) not in (list,tuple):
        keys = [keys]
    for key in keys:
        find = 1
        if type(maps)==dict:
            find = key in maps
        elif type(maps) in (list, tuple):
            try:
                key = int(key)
            except:
                key = len(maps)
            find = key<len(maps)
        else:
            find = False
        if not find:
            return None, find
        maps = maps[key]
    return maps, True

pass
def l2m(arr, **maps):
    rst = {}
    i = 0
    for k in maps:
        if i<len(arr):
            val = arr[i]
        else:
            val = maps[k]
        rst[k] = val
        i+=1
    return rst

pass
def deep_fill_argx(src, target, replace=1):
    # print(f"src:{src}")
    # print(f"target:{target}")
    if not isinstance(src, Args):
        if type(src) in (list, tuple):
            src = Args(src)
        elif type(src)==dict:
            src = Args([], src)
        else:
            return
    tp = type(target)
    if not isinstance(target, Args):
        if tp == list:
            target = Args(target)
        elif tp == dict:
            target = Args([], target)
        else:
            return
    for it in src.lists:
        target.args.append(it)
    for k,v in src.dicts.items():
        if k not in target.dicts:
            target.dicts[k] = v
            continue
        tv = target.dicts[k]
        if Args.is_collect(tv) and Args.is_collect(v):
            deep_fill_argx(v, tv, replace)
        elif replace:
            #print(f"[TESTZ] replace: {k}:{tv} to {v}")
            target.dicts[k] = v



def deep_update(target, src, replace=1):
    """
        dict深层更新，src[key]是dict就深入更新，否则:
            src有而target没有就替换，否则：
                replace=1就替换
    """
    for k in src:
        val = src[k]
        if k not in target:
            target[k] = val
            continue
        mval = target[k]
        if type(mval) == dict and type(val)==dict:
            deep_update(mval, val, replace)
        else:
            if replace:
                target[k] = val

pass
update = deep_update
def deep_fill(src, target, replace=1):
    return deep_update(target, src, replace)

pass
def deep_clone(obj):
    if type(obj)==list:
        rst = []
        for it in obj:
            rst.append(deep_clone(it))
    elif type(obj)==dict:
        rst = {}
        for k,v in obj.items():
            rst[k] = deep_clone(v)
    else:
        rst = obj
    return rst
pass
fill=deep_fill

def maps(**kv):
    return kv

pass

def flush_maps(maps, fc_key = lambda x:x.split(".") if type(x)==str else [x], visit_list=False):
    if type(maps)==list:
        if not visit_list:
            return maps
        rst = [flush_maps(it, fc_key, visit_list) for it in maps]
        return rst
    if type(maps)!=dict:
        return maps
    rst = {}
    for k,v in maps.items():
        ks = fc_key(k)
        v = flush_maps(v, fc_key, visit_list)
        dset(rst, ks, v)
    return rst

pass
def unflush_maps_item(rst, keys, val, fc_key):
    if type(val)!=dict:
        key = fc_key(keys)
        rst[key] = val
        return
    for k, v in val.items():
        unflush_maps_item(rst, keys+[k], v, fc_key)
def unflush_maps(maps, fc_key = lambda x:".".join([str(k) for k in x]) if type(x) in (list, tuple) else str(x)):
    if type(maps)!=dict:
        return maps
    rst = {}
    for k,v in maps.items():
        unflush_maps_item(rst, [k], v, fc_key)
    return rst



from .. import base
from .. import item
from .. import exp
from . import lr
class MapDeal(lr.LRDeal):
    """
        分隔符，有分隔符后将缓存的数据当作字符串
    """
    def init(self, left, right):
        self.sp().init(left, right, "map")
    def build(self, arr, l_item):
        rst = []
        for _item in arr:
            if not _item.check(is_keyval = 1):
                print(arr)
                print("error _item:", _item)
                raise exp.FormatExp("an not key-val item find in map", _item.pos, _item.val)
            rst.append(_item.val)
        maps = {k[0]:k[1] for k in rst}
        return item.DealItem(maps, l_item.pos, self.id(), is_val = 1, is_map = 1)

pass



from .writerz import mg, itemz, base, conf
from .writerz.deal import listz, mapz, strz, reval, jsonval, listmapz
from . import file
from decimal import Decimal
from .base import Args
pts = [
    "[\+\-]?\d+",
    "[\+\-]?\d+\.\d+",
    "[\+\-]?\d+e[\+\-]?\d+",
    "null",
    "true",
    "false",
    "[\s\S]*[\n\r\t\:\[\]\{\}\(\)\,\:\=\'\<\>\" \|\#\;\/][\s\S]*"
]
def build(json_format=False, args = False):
    mgs = mg.Manager()
    if not json_format:
        mgs.add([bytes, str], strz.StrDeal('"','"', pts))
        mgs.add(float, reval.ValDeal(lambda x:str(x)))
        mgs.add(int, reval.ValDeal(lambda x:str(x)))
        mgs.add(type(None), reval.ValDeal(lambda x:'null'))
        mgs.add(Decimal, reval.ValDeal(lambda x:str(x)))
        mgs.add(bool, reval.ValDeal(lambda x:'true' if x else 'false'))
    else:
        mgs.add([bytes, str, float, int, type(None), Decimal, bool], jsonval.ValDeal())
    mgs.add([list, tuple, set], listz.ListDeal('[',']',','))
    mgs.add(dict, mapz.MapDeal('{','}',':',','))
    if args:
        mgs.add(Args, listmapz.ListMapDeal('(',')','=',','))
    return mgs

pass
def dumps(obj, bytes = 0, format = 0, not_format_height = 1, json_format= 0, args = False, deep=None):
    if deep is not None:
        not_format_height = deep+1
    cf = conf.Conf()
    cf.set(bytes=bytes, format=format, nfmt_height=not_format_height)
    if format:
        cf.set(set=1, prev=1,line=4, spc=' ')
    else:
        cf.set(set=1, prev=1)
    mgs = build(json_format, args)
    return mgs.dump(obj, cf)

pass
def dumpx(*a, **b):
    return dumps(*a, **b, args=True)

pass

def dumpf(obj, filepath, bytes = 0, format = 0, not_format_height = 1, json_format= 0, mode = 'wb', args = False, deep=None):
    if deep is not None:
        not_format_height = deep+1
    s = dumps(obj, bytes = bytes, format = format, not_format_height = not_format_height, json_format= json_format).encode("utf-8")
    file.fwrite(s, filepath, mode)

pass
def dumpxf(*a, **b):
    return dumpf(*a, **b, args=True)

pass

def dump(output, obj, *argv, **maps):
    rs = dumps(obj, *argv, **maps)
    output(rs)

pass



#

import sys
from buildz.gpuz.torch import DictCache
from buildz.gpuz.test import analyze
from buildz import pyz
import torch,time
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
cpu,cuda = analyze.dvs
class Model(nn.Module):
    def __init__(self, dims, num):
        super().__init__()
        nets=[]
        for i in range(num):
            nets.append(nn.Linear(dims,dims))
            nets.append(nn.LeakyReLU())
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
    def size(self):
        return sum([analyze.unit_sz(net) for net in self.nets])

pass
class TestDataset(Dataset):
    def __init__(self, n, dims):
        self.n = n
        self.dims = dims
        self.datas = torch.rand(n, dims)
        sz = analyze.sz(self.datas)
        sz, unit = analyze.show_size(sz)
        print(f"data size: {sz} {unit}")
    def __len__(self):
        return self.n
    def __getitem__(self, i):
        return self.datas[i], self.datas[i]

pass

def test():
    '''
        这些参数是不用DictCache的时候8GB显存刚好能跑
        测试显存不够用的情况，要么用其他程序占显存，要么改参数，num改大一些显存就不够用了
        也提供了个占用显存的简单代码:
            占用3GB显存：
                python -m buildz.gpuz.test.take_gpu_mem 12
            占用4GB显存：
                python -m buildz.gpuz.test.take_gpu_mem 20

    '''
    nets=10
    dims=2000
    loop = 5
    datas = 60
    batch=30
    lr=0.0001
    win_size=3
    args = sys.argv[1:]
    mark_train = True
    if len(args)>0:
        mark_train = args.pop(0).lower()=='train'
    modes = 'cuda,cache,cpu'
    if len(args)>0:
        modes = args.pop(0)
    num = 12
    if len(args)>0:
        num = int(args.pop(0))
    print(f"num: {num}")
    def fc_gen():
        mds = [Model(dims, nets) for i in range(num)]
        mds_sz = [md.size() for md in mds]
        sz, unit = analyze.show_size(sum(mds_sz))
        print(f"Model Size: {sz} {unit}")
        opts =[optim.Adam(md.parameters(), lr=lr) for md in mds]
        gmodel = nn.Sequential(*mds)
        gopt = optim.Adam(gmodel.parameters(), lr=lr)
        return mds, gmodel, opts, gopt
    def fc_opt(net, opt):
        torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
        opt.step()
    ds = TestDataset(datas, dims)
    dl = DataLoader(ds, batch)
    loss_fn = torch.nn.MSELoss()
    analyze.analyzes(mark_train, loop, fc_gen, dl, loss_fn, fc_opt, win_size, modes)

pyz.lc(locals(),test)


from .. import base
from .. import item
from .. import exp
from . import lr
from . import reval
class Fcs:
    def __init__(self):
        self.maps = {}
    def set(self, _type, fc):
        self.maps[_type] = fc
    def __call__(self, val, _type):
        if _type not in self.maps:
            raise Exception("unreginize type:"+_type)
        return self.maps[_type](val)

pass


class LRValDeal(lr.LRDeal):
    """
    """
    def init(self, left, right, fc):
        self.sp().init(left, right, 'lrval')
        self.fc = fc
    def build(self, arr, l_item):
        rst = []
        if len(arr)!=2:
            raise exp.FormatExp("format error find in lrval:", l_item.pos, l_item.val)
        for _item in arr:
            if not _item.check(is_val = 1):
                raise exp.FormatExp("format error find in list:", _item.pos, _item.val)
            rst.append(_item.val)
        try:
            val = self.fc(rst[0], rst[1])
        except Exception as exp1:
            print("exp:", exp1)
            raise exp.FormatExp("error in lrval fc:\""+str(exp1)+"\"", arr[0].pos, arr[0].val)
        return item.DealItem(val, l_item.pos, self.id(), is_val = 1, is_lrval = 1)

pass

class LRReValDeal(reval.ValDeal):
    """
    """
    def efc(self, val):
        arr = val.split(self.like(self.spt, val))
        val, _type = arr
        return self._fc(val, _type)
    def init(self, spt, fc):
        self.spt = spt
        self._fc = fc
        pt = "^.+\s*\\"+spt+"\s*.+$"
        self.sp().init(pt, self.efc)

pass




#print("xf.__init__")
from .readz import loads, load, loadf, loads_args, loadx, is_args, loadxf
from .writez import dumps, dump, dumpf, dumpx, dumpxf
from .file import *
from .mapz import *
from .xargs import fetch as args

try:
    # C++加速代码
    #from .cpp import pcxf
    #print(f"pcxf:", pcxf)
    #print(f"pcxf.loads:", dir(pcxf))
    from buildz.xf.cpp.pcxf import loads, loadx
    loads_args = loadx
except Exception as exp:
    #print("init not cpp:",exp)
    pass
pass
__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"



#coding=utf-8
from ..ioc.base import Base, EncapeData, IdNotFoundError
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
import re
class RefsDeal(FormatDeal):
    """
        引用refs，返回的是一个列表，长度可能为0:
            {
                id: id
                type: refs
                refs|key: 引导数据id正则匹配表达式
                info: item_conf, 额外的引用信息, 默认null
            }
        简写:
            [[refs, id], [key, force_new], info]
            [[refs, id], key, info]
        极简:
            [refs, key]
        例:
            [refs, r".*obj\.test.*"] // 数据项"obj.test"的引用
    """
    def init(self, fp_lists=None, fp_defaults=None):
        super().init("RefsDeal", fp_lists, fp_defaults, join(dp, "conf", "refs_lists.js"), None)
    def match(self, key, pt):
        rst = len(re.findall(pt, key))>0
        # if rst:
        #     print(f"[TESTZ] refs match: {key} in {pt}")
        # else:
        #     print(f"[TESTZ] refs NOT match: {key} in {pt}")
        return rst
    def deal(self, edata:EncapeData):
        data = edata.data
        data = self.fill(data)
        pt = xf.get_first(data, 'refs', 'key')
        info = xf.g(data, info=None)
        if info is not None and type(info)==dict:
            #info = {k:self.get_obj(info, edata.conf, src = edata.src) for k in info}
            info = {'type':'map', 'data':info}
            info = self.get_obj(info, edata.conf, src = edata.src)
        else:
            info = {}
        _info = edata.info
        if type(_info)==dict:
            xf.deep_update(info, _info, 1)
        vars = edata.conf.var_keys()
        vars = [key for key in vars if self.match(key, pt)]
        rst = [edata.conf.get_var(key)[0] for key in vars]
        ids = edata.confs.full_ids(edata.conf)
        ids = [obj for obj in ids if self.match(obj[0], pt) or self.match(obj[1], pt)]
        objs = []
        for obj in ids:
            #print(f"[TESTZ] REFS get: {obj[2]}.get({obj[1]}, src={edata.src})")
            _obj = obj[2].get(obj[1], info = info, src = edata.src)
            objs.append(_obj)
        #objs = [obj[2].get(obj[1], info = info, src = edata.src) for obj in ids]
        rst+=objs
        return rst

pass



#

from .pool import *
import threading
class FcTask(Task):
    def init(self, fc, args=[], maps={}, fc_exp = None):
        self.fc = fc
        self.args = args
        self.maps = maps
        self.fc_exp = fc_exp
    def run(self):
        return self.fc(*self.args, **self.maps)
    def deal_exp(self, exp, traceback_exp):
        if self.fc_exp is not None:
            return self.fc_exp(exp, traceback_exp)
        print("FcTask exp:", exp)
        print("traceback:", traceback_exp)
class SimpleTasks(Tasks):
    def init(self, wait_sec = 0.1):
        self.condition = threading.Condition()
        self.tasks = []
        self.wait_sec = wait_sec
        self.wait_count=0
    def waits(self):
        return len(self.tasks)
    def add(self, task):
        with self.condition:
            self.tasks.append(task)
            if self.wait_count>0:
                self.condition.notify(1)
    def pop(self):
        with self.condition:
            if len(self.tasks)==0:
                self.wait_count+=1
                self.condition.wait(self.wait_sec)
                self.wait_count-=1
            if len(self.tasks)==0:
                return None
            return self.tasks.pop(0)
class RangeManager(WorkerManager):
    def init(self, vmin=0, vmax=-1):
        self.min = vmin
        self.max = vmax
    def change(self, workers, tasks, first_call):
        waits = tasks.waits()
        curr = len(workers)
        if curr<self.min:
            return self.min-curr
        if self.max>0 and curr>self.max:
            return self.max-curr
        frees = len([wk for wk in workers if wk.wait_count>1])
        needs = waits-frees
        rst = curr+needs
        if needs==0:
            return 0
        if needs>0:
            if self.max>0 and rst>self.max:
                needs = self.max-curr
            return needs
        else:
            if rst<self.min:
                needs = self.min-curr
            return needs

pass            



from html.parser import HTMLParser
from .. import xf
from ..base import Base
import re
class SearchResult(Base):
    def str(self):
        return str(self.arr)
    def init(self, arr):
        self.arr = arr
        self.size = len(arr)
    def run(self, fc_name, *a, **b):
        rst = []
        for it in self.arr:
            fc = getattr(it, fc_name)
            rst+=fc(*a,**b).data()
        return SearchResult(rst)
    def searchs(self, *args, **maps):
        return self.run("searchs", *args, **maps)
    def finds(self, s):
        return self.run("finds", s)
    def tags(self, _tag):
        return self.searchs(tag=_tag)
    def call(self):
        return self.data()
    def data(self):
        return self.arr
    def text(self):
        return [it.text for it in self.arr]
    def texts(self):
        return [it.texts for it in self.arr]
    def tag(self):
        return [it.tag for it in self.arr]
    def gkey(self, attrs, k):
        if k is None:
            return attrs
        if k in attrs:
            return attrs[k]
        return None
    def attrs(self, k=None):
        return [self.gkey(it.attrs,k) for it in self.arr]

pass
class HtmlTag:
    def data(self):
        return self.to_maps()
    def to_maps(self):
        nodes = [n.to_maps() for n in self.nodes]
        rst = {'tag': self.tag, 'attrs': self.attrs, 'texts': self.texts, 'text': self.text, 'nodes': nodes}
        return rst
    def __str__(self):
        return xf.dumps(self.to_maps())
    def __repr__(self):
        return self.__str__()
    def match(self, val, pt):
        tag = self.tag
        text = self.text
        texts = self.texts
        attrs = self.attrs
        if type(pt)==list:
            tp = pt[0]
            v = pt[1]
        else:
            tp = "="
            v = pt
        if val is None and tp not in ["eval", "exec"]:
            return False
        if tp == '>':
            return val>v
        elif tp == "<":
            return val<v
        elif tp == "=":
            return val == v
        elif tp == ">=":
            return val>=v
        elif tp=="<=":
            return val<=v
        elif tp == "re":
            return len(re.findall(v, val))>0
        elif tp == 'eval':
            return eval(v)
        elif tp == 'exec':
            exec(v)
            return self.val
        else:
            raise Exception(f"not impl match type: '{tp}'")
    def get(self, key):
        ks = key.split(".")
        v = ks[0]
        if not hasattr(self, v):
            return None
        obj = getattr(self, v)
        for k in ks[1:]:
            if type(obj)==dict:
                if k not in obj:
                    return None
                obj = obj[k]
            elif type(obj)==list:
                k = int(k)
                if k>=len(obj):
                    return None
                obj = obj[k]
            else:
                return None
        return obj
    def check(self, args, maps):
        arr = args+list(maps.items())
        for k,v in arr:
            if k == "$":
                if not self.match(None, ["eval", v]):
                    return False
            else:
                val = self.get(k)
                if not self.match(val, v):
                    return False
        return True
    def __init__(self, tag, attrs=None):
        self.tag = tag
        if attrs is None:
            attrs = {}
        self.attrs = attrs
        self.nodes = []
        self.text = None
        self.texts = []
    def _search(self, rst, args, maps):
        if self.check(args, maps):
            rst.append(self)
        for n in self.nodes:
            n._search(rst, args, maps)
    def searchs(self, *args, **maps):
        tmp = []
        for k in args:
            if type(k)!=list:
                k = xf.loads(k)
            tmp.append(k)
        rst = []
        self._search(rst, tmp, maps)
        return SearchResult(rst)
    def tags(self, _tag):
        return SearchResult(self.searchs(tag = _tag))
    def finds(self, s):
        rst = []
        args = []
        maps = {}
        dt = xf.loads(s)
        if type(dt)==list:
            args = dt
        else:
            maps = dt
        if len(args)==2 and type(args[0])!=list:
            args = [args]
        self._search(rst, args, maps)
        return SearchResult(rst)
    def add_node(self, node):
        self.nodes.append(node)
    def add_text(self, text):
        self.text = text
        self.texts.append(text)

pass

class MyHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.data = HtmlTag("document")
        self.stacks = [self.data]
    def handle_comment(self, data):
        "处理注释，< !-- -->之间的文本"
        pass
    def handle_startendtag(self, tag, attrs):
        "处理自己结束的标签，如< img />"
        self.handle_starttag(tag, attrs)
        self.handle_endtag(tag)
        pass
    def handle_starttag(self, tag, attrs):
        "处理开始标签，比如< div>；这里的attrs获取到的是属性列表，属性以元组的方式展示"
        attrs = {k:v for k,v in attrs}
        tag = HtmlTag(tag, attrs)
        self.stacks[-1].add_node(tag)
        self.stacks.append(tag)
        #print(f"Encountered a {tag} start tag")
        #for attr, value in attrs:
        #    print(f"   {attr} = {value}")
    def handle_data(self, data):
        self.stacks[-1].add_text(data)
        "处理数据，标签之间的文本"
        #print(f"----handle data in tags: {data}")
    def handle_endtag(self, tag):
        self.stacks.pop(-1)
        "处理结束标签,比如< /div>"
        #print(f"Encountered a {tag} end tag")

pass

def parse(text):
    obj = MyHTMLParser()
    obj.feed(text)
    return obj.data

pass


#
from ..ioc.base import Base, EncapeData
from .base import FormatData,FormatDeal
from ..ioc.decorator import decorator,IOCDObj
from buildz import xf, pyz
from buildz import Base as Basez
import os
dp = os.path.dirname(__file__)
join = os.path.join
class VarDeal(FormatDeal):
    """
        代码变量var:
            {
                id:id
                type: var
                var|data: string
            }
        简写:
            [[var, id], data]
            [var, data]
        例:
            [var, buildz.pyz.is_windows] // 返回buildz.pyz下的is_windows
    """
    def init(self, fp_lists = None, fp_defaults = None):
        self.singles = {}
        self.sources = {}
        super().init("VarDeal", fp_lists, fp_defaults, 
            join(dp, "conf", "var_lists.js"))
    def deal(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        conf = edata.conf
        data = self.format(data)
        src = edata.src
        key = xf.get_first(data, "var", "key", "data")
        key = pyz.load(key)
        return key

pass

class IOCVar(IOCDObj):
    def init(self, id):
        super().init()
        self.id = id
    def call(self, cls):
        src = cls.__module__+"."+cls.__name__
        conf = {}
        conf['id'] = self.id
        conf['type'] = 'var'
        conf['var'] = src
        self.decorator.add_datas(conf)
        return cls

pass
decorator.regist("var", IOCVar)




import os
import sys
dp = os.path.dirname(__file__)
gdp = os.path.join(dp, "lib")

def init_path(dp):
    path = os.environ["PATH"] 
    if sys.platform.lower().find("win")>=0:
        cb = ";"
    else:
        cb = ":"
    path = path +cb+dp
    os.environ["PATH"] = path

pass

#init_path(dp)
import sys
from .basez import SimpleDv, fetch
from .structz import CMD
class Db(SimpleDv):
    # func to impl:
    def to_list(self, rst):
        rows = self.cursor.description
        keys = [k[0].lower() for k in rows]
        result = []
        result.append(keys)
        if rst is None or len(rst)==0:
            return result
        result += rst
        return result
    def new_con(self):
        if self.port is not None:
            add = ":"+str(self.port)
        else:
            add = ""
        try:
            import cx_Oracle as pymysql
        except ModuleNotFoundError:
            raise Exception("module not found, try: pip install cx-Oracle")
        return pymysql.connect(self.user, self.pwd, self.host+add+"/"+self.db)
    def new_cursor(self):
        return self.con.cursor()
    def init(self, *argv, **maps):
        pass
    def sql_tables(self, table = None):
        """
            if not note, use name instead
            require:
                table_name, table_note
        """
        query_table = ""
        if table is not None:
            query_table =  f" and t.table_name = '{table.upper()}'"
        return f"select t.table_name as table_name, tc.comments as table_note from all_tables t inner join all_tab_comments tc on t.table_name=tc.table_name where t.owner='{self.user.upper()}' {query_table}"
    def sql_columns(self, table):
        """
            require:
                table_name, column_name, column_type, column_default, nullable, column_offset, column_note
        """
        return f"select c.table_name as table_name, c.column_name as column_name, c.data_type|| '(' ||c.data_length||','||c.data_precision||','||c.data_scale||')' as column_type, c.data_default as column_default, c.nullable as nullable, c.column_id as column_offset, cc.comments as column_note from all_tab_columns c inner join all_col_comments cc on c.column_name=cc.column_name and c.table_name=cc.table_name where c.table_name='{table.upper()}' order by c.column_id asc"
    def sql_indexes(self, table=None, index=None):
        """
            require:
                table_name, index_name, is_unique, index_type, index_note
        """
        query_table = ""
        if table is not None:
            query_table =  f" and table_name = '{table.upper()}'"
            if index is not None:
                query_table+= f" and index_name='{index.upper()}'"
        return f"select table_name, index_name,  case uniqueness when 'UNIQUE' then 1 else 0 end as is_unique, index_type, '' as index_note from all_indexes where owner = '{self.user.upper()}'{query_table}"
    def sql_index_keys(self, table, index):
        """
            require:
                table_name, index_name, column_name, index_offset, column_note
        """
        return f"select table_name, index_name, column_name, column_position as index_offset, '' as note from all_ind_columns where index_name='{index.upper()}' and table_name='{table.upper()}' order by column_position asc"

pass
def build(argv, conf):
    k = 'oracle_lib'
    dp = gdp
    if k in conf:
        dp = conf[k]
    init_path(dp)
    #print(f"oracle lib: {dp}")
    dv = Db(*fetch(argv))
    return dv
def buildbk(argv, conf):
    return CMD(make(argv, conf))

pass




from buildz import Base
import os,time, traceback
class FpsListener(Base):
    def init(self, wait_sec=0.1):
        self.fps = {}
        self.wait_sec = wait_sec
        self.running=False
    def set_wait(self, wait):
        self.wait_sec = wait
    def update(self, fps):
        pass
    def deal_exp(self, exp, format_exc):
        pass
    def add(self, fp, last_update=False):
        if type(last_update)==bool:
            if last_update:
                last_update = os.path.getmtime(fp)
            else:
                last_update = 0
        self.fps[fp] = last_update
    def clean(self):
        self.fps = {}
    def reset(self, fps, last_update=False):
        self.clean()
        for fp in fps:
            self.add(fp, last_update)
    def run(self):
        self.running = True
        while self.running:
            self.work()
            time.sleep(self.wait_sec)
    def work(self):
        upds = []
        for fp,sec in self.fps.items():
            curr = os.path.getmtime(fp)
            if curr!=sec:
                upds.append(fp)
            self.fps[fp]=curr
        if len(upds)==0:
            return
        try:
            self.update(upds)
        except Exception as exp:
            trs = traceback.format_exc()
            try:
                self.deal_exp(exp, trs)
            except Exception as exp1:
                print(f"exp in deal_exp({exp}): {exp1}")
                traceback.print_exc()

pass

class FcFpsListener(FpsListener):
    def init(self,wait_sec=0.1):
        super().init(wait_sec)
        self.fc_update = None
        self.fc_exp = None
    def set_update(self, fc):
        self.fc_update = fc
    def set_deal_exp(self, fc):
        self.fc_exp = fc
    def update(self, fps):
        return self.fc_update(fps)
    def deal_exp(self, exp, fmt_exc):
        return self.fc_exp(exp, fmt_exc)

pass


#

from .lst import FcFpsListener
from .. import Base, xf, dz, pathz, pyz
import os, sys
from multiprocessing import Process
class TestFc:
    def __init__(self, fc):
        self.fc = fc
    def __call__(self, *a, **b):
        print(f"[TESTZ] call: {self.fc}(*{a}, **{b})")
        self.fc(*a,**b)
class Test:
    def __init__(self, file):
        self.file = file
    # def write(self):
    #     pass
    def __getattr__(self, key):
        fc = getattr(self.file, key)
        print(f"[TESTZ] get: {key} = {fc}")
        if not callable:
            return fc
        #return getattr(self.file, key)
        return TestFc(fc)
class Runner(Base):
    def init(self, fp, lst):
        fp = os.path.abspath(fp)
        self.dp = os.path.dirname(fp)
        self.path = pathz.Path(fp=self.dp)
        self.fp = fp
        self.lst = lst
        self.lst.set_update(self.update)
        self.lst.set_deal_exp(self.deal_exp)
        self.process = None
        self.reset()
    def reset(self,last_update=False):
        conf = xf.loadf(self.fp)
        fps, target, dp = dz.g(conf, fps = [], run=None, dp="")
        dp = self.path.fp(dp)
        sys.path.append(dp)
        self.path.set("sc", dp)
        fps = [self.path.sc(fp) for fp in fps]
        self.lst.clean()
        self.lst.add(self.fp, last_update)
        [self.lst.add(fp, last_update) for fp in fps]
        self.target=target
        self.conf = conf
    @staticmethod
    def process_update(conf):
        target, sc_path, args, stdin, stdout, stderr = dz.g(conf, run=None,sc_path=__file__, args=[], stdin = "std", stdout = "std", stderr="std")
        encoding = dz.g(conf, encoding="utf-8")
        import sys, os
        args = [sc_path]+args
        sys.argv = args
        stds = [stdin, stdout, stderr]
        stds = [list(k) if type(k) in (list, tuple) else [k] for k in stds]
        modes = "rww"
        for std, mode in zip(stds, modes):
            if len(std)<3:
                std.append(mode)
        objs = []
        files = {}
        for std in stds:
            do_close = False
            if std[0]=='std':
                obj = sys.stdin if std[-1]=='r' else sys.stdout
            elif std[0]=='str':
                obj = io.StringIO(std[1])
            else:
                if std[1] not in files:
                    files[std[1]] = open(std[1], std[-1], encoding=encoding)
                obj = files[std[1]]
            objs.append(obj)
        #bak = [sys.stdin, sys.stdout, sys.stderr]
        sys.stdin = objs[0]
        sys.stdout = objs[1]
        sys.stderr = objs[2]#Test(objs[2])#objs[2]
        #print(f"stdin: {sys.stdin}, stdout: {sys.stdout}, stderr: {sys.stderr}")
        #print(f"files: {files}")
        try:
            #print(f"start process: {os.getpid()}")
            fc = pyz.load(target)
            if callable(fc):
                fc()
        except Exception as exp:
            #print(f"exp:", exp)
            #print(f"exp in process: {os.getpid()}")
            raise exp
        finally:
            #print(f"done process: {os.getpid()}, {objs[2].closed}")
            for fp, f in files.items():
                f.flush()
            #sys.stdout = bak[1]
    def update(self, fps):
        if self.process is not None:
            if self.process.is_alive():
                self.process.kill()
            self.process = None
        self.reset(True)
        p = Process(target = self.process_update,args=[self.conf], daemon=True)
        p.start()
        self.process = p
    def deal_exp(self, exp, fmt_exc):
        print(f"exp: {exp}")
        print(f"traceback: \n{fmt_exc}")

def test():
    fp = sys.argv[1]
    lst = FcFpsListener()
    runner = Runner(fp, lst)
    lst.run()

pass
pyz.lc(locals(), test)


#

from buildz import evalz, xf
class VarEval(evalz.Eval):
    def init(self, key):
        self.key = key
    def call(self, obj):
        return obj[self.key]
class VarBuild(evalz.Build):
    def call(self, data):
        key = data['data'][0]
        return VarEval(key)
vb = VarBuild()
builder = evalz.EvalBuilder.Make().set("v", vb).set("var", vb)

conf = xf.loads(r"""
&, 
['=', [v, name], test]
['=', [v, value], 1024]
""")

eval = builder(conf)
data = xf.loads(r"""
name=test
value = 1024
""")
print(eval(data))

"""
python -m buildz.evalz.test

"""


#
from .tdata import TagData
from .datas import Datas
from .dataset import Dataset
class Confs(Datas):
    @staticmethod
    def is_conf(obj):
        return type(obj) in (list, tuple, dict)
    def init(self, ns=None, id=None, deal_ns = None, dts=None):
        super().init(ns, id, dts)
        self.deal_ns = deal_ns
    def set(self, key, val, tag=None):
        val = (val, self.deal_ns, self.id)
        super().set(key, val, tag)
        




from .base import *
from ... import dz,pyz
from .val import ValEncape
class IOCDeal(BaseDeal):
    def deal(self, conf, unit):
        target = dz.g(conf, target='unit')
        data = None
        if target == 'unit':
            data = unit
        elif target =='conf':
            data = conf
        elif target in ('manager','mg'):
            data = unit.mg
        assert data is not None, f"[IOC] unknown target '{target}' in {conf}"
        return ValEncape(data)

pass


#

from .ioc import confs
from .ioc.confs import ConfsList, Confs
from .ioc_deal.base import default_deals
from . import wrap
def build(fp_init = None, add_default_deals = True, add_wrap = True):
    obj = confs.Confs()
    if (fp_init is not None):
        if type(fp_init)==dict:
            obj.init(fp_init)
        else:
            obj.init_fp(fp_init)
    if add_default_deals:
        obj.add_fp(default_deals)
    if add_wrap:
        obj.add_decorator()
    return obj

pass


'''
格式化报错
'''
class FormatExpBak(Exception):
    def __init__(self, err, data, s = ""):
        if len(s)==0:
            errs = "Error: {err}, line: {line}, index: {index}".format(err = err, line = data[0], index = data[1])
        else:
            errs = "Error: {err}, line: {line}, index: {index}, content: [{s}]".format(err = err, line = data[0], index = data[1], s = s)
        super(FormatExp, self).__init__(errs)

pass

class Exp(Exception):
    def __init__(self, msg, pos):
        msg = f"{msg}, [OFFSET]: {pos}"
        self.pos = pos
        super().__init__(msg)

pass

def deal(exp, buff):
    pos_str = buff.pos2str(exp.pos)
    s = str(exp)
    msg = f"{s} [CONTENT]: '{pos_str}'"
    raise Exception(msg)


#

from .base import *
from ..ioc_deal import obj

class ObjectDeal(obj.ObjectDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'source', need=1)
        conf.index(2, 'args')
        conf.index(3, 'maps')
        conf.index(4, 'sets')
        conf.index(5, 'after_set')
        conf.key('source', 'src'.split(","), need=1)
        conf.key('before_set', 'before_call'.split(","))
        conf.key('after_set', 'after_call,call'.split(","))
        self.update = conf

pass


#
import datetime,decimal
def lower(val):
    rst = ""
    for c in val:
        if c.lower()!=c:
            c = "_"+c.lower()
        rst+=c
    return rst
pass
def upper(val):
    rst = ""
    mark_up = False
    for c in val:
        if c == "_":
            mark_up = True
            continue
        if mark_up:
            c = c.upper()
            mark_up = False
        rst+=c
    return rst
pass
def py2sql(val):
    if val is None:
        return 'null'
    if type(val)==str:
        return f"'{val}'"
    if type(val) == datetime.datetime:
        val = str(val) 
    if type(val) == decimal.Decimal:
        val = str(val)
    return val
pass


#coding=utf-8
from . import runz
from .. import Base, xf, pyz


from ..sc.subrun import Runner
from ..sc.lst import FcFpsListener
from buildz.db import runz
from buildz import dz
class ScRunner(Runner):
    def loadf(self, fp):
        conf = xf.loadf(fp)
        dz.s(conf, fps=[dz.g(conf, src=None)])
        return conf
    def process_command(self):
        cmd = f"python -m buildz.db.subsc {self.fp}"
        return cmd.split(" ")
class DbRunner(Base):
    def init(self, conf_fp, listener):
        self.fp = conf_fp
        self.lstn = listener
        self.lstn.set_update(self.update)
        self.lstn.set_deal_exp(self.deal_exp)
        self.reset()
    def reset(self, last_update=False):
        self.lstn.clean()
        self.lstn.add(self.fp,last_update)
        conf = xf.loadf(self.fp)
        src = xf.g(conf, src=None)
        self.lstn.add(src,last_update)
        sec = xf.g(conf, sec=0.1)
        self.lstn.set_wait(sec)
    def update(self, fps):
        self.reset(True)
        runz.test(self.fp)
    def deal_exp(self, exp, fmt_exc):
        print(f"exp: {exp}")
        print(f"traceback: \n{fmt_exc}")

pass
#runz.main()
def run():
    import sys
    fp = runz.FP
    if len(sys.argv)>1:
        fp = sys.argv[1]
    lst = FcFpsListener()
    runner = DbRunner(fp, lst)
    lst.run()

pass
def test():
    import sys
    fp = runz.FP
    if len(sys.argv)>1:
        fp = sys.argv[1]
    lst = FcFpsListener()
    runner = ScRunner(fp, lst)
    lst.run()

pyz.lc(locals(), test)





from .base import *
from ... import dz,pyz
from ..ioc.single import Single
class RefEncape(BaseEncape):
    '''
    '''
    def init(self, ref, default, unit, env=-1):
        super().init()
        self.unit = unit
        self.ref, self.default = ref, default
        self.env = env
    def call(self, params=None, **maps):
        ref = self.ref
        if isinstance(ref, Encape):
            ref = ref(params)
        if self.env==0:
            obj, _, find = self.unit.get_env(ref)
            if find:
                return obj
        obj, find = self.unit.get(ref)
        if not find:
            if self.env==1:
                obj, _, find = self.unit.get_env(ref)
                if find:
                    return obj
            assert self.default[0], f"ref '{ref}' not found"
            val = self.default[1]
            if isinstance(val, Encape):
                val = val(params)
            obj = val
        return obj
class RefDeal(BaseDeal):
    def init(self, env = -1):
        super().init()
        self.env = env
    def deal(self, conf, unit):
        assert 'ref' in conf, f"[REF] key 'ref' not found in {conf}"
        ref = dz.g(conf, ref=None)
        if 'default' in conf:
            default = [1,conf['default']]
        else:
            default = [0,None]
        ref = self.get_encape(ref, unit)
        if default[0]:
            default[1] = self.get_encape(default[1], unit)
        env = None
        if 'env' in conf:
            env = conf['env']
        if env is None:
            env = self.env
        encape = RefEncape(ref, default, unit, env)
        return encape

pass